/** @file display.cpp

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ display ]
*/

#include "display.h"
#include "../../fileio.h"
#include "../../config.h"

#define CURSOR_MAXPOS 0xffffff

#ifdef _DEBUG
//#define DEBUG_CRTMON
#endif

#define COLOR16A(bb,rr,gg,ii) color16[(bb) | ((rr) << 1) | ((gg) << 2) | ((ii) << 3)]
#define COLOR64A(bb0,rr0,gg0,ii0,bb1,rr1,gg1,ii1) color64[(bb0) | ((rr0) << 1) | ((gg0) << 2) | ((ii0) << 3)][(bb1) | ((rr1) << 1) | ((gg1) << 2) | ((ii1) << 3)]

void DISPLAY::initialize()
{
	memset(font, 0, sizeof(font));
	font_rom_loaded = false;
	l3font_rom_loaded = false;
	font_rom_loaded_at_first = false;

	vm_pause = emu->get_pause_ptr();

	// load rom images
	load_font_rom_file();

	emu->get_rgbaformat(&Rmask, &Gmask, &Bmask, &Amask, &Rshift, &Gshift, &Bshift, &Ashift);

	// create pc palette
	for(int i = 0; i < 2; i++) {
		for(int g = 0; g < 2; g++) {
			for(int r = 0; r < 2; r++) {
				for(int b = 0; b < 2; b++) {
					int bb = b * (i ? 0xff : 0x7f);
					int rr = r * (i ? 0xff : 0x7f);
					int gg = g * (i ? 0xff : 0x7f);
					COLOR16A(b, r, g, i) = emu->map_rgbcolor(rr,gg,bb);
				}
			}
		}
	}
	const uint8 mm[16] = {
		0x00, 0x00, 0x00, 0x00,
		0x1f, 0x3f, 0x1f, 0x3f,
		0x5f, 0x5f, 0x7f, 0x7f,
		0x9f, 0xbf, 0xdf, 0xff
	};
	for(int i = 0; i < 4; i++) {
		for(int g = 0; g < 4; g++) {
			for(int r = 0; r < 4; r++) {
				for(int b = 0; b < 4; b++) {
					int bb = mm[b * 4 + i];
					int rr = mm[r * 4 + i];
					int gg = mm[g * 4 + i];
					if (bb < 0) bb = 0;
					if (rr < 0) rr = 0;
					if (gg < 0) gg = 0;
					if (bb > 0xff) bb = 0xff;
					if (rr > 0xff) rr = 0xff;
					if (gg > 0xff) gg = 0xff;
					COLOR64A(b >> 1, r >> 1, g >> 1, i >> 1, b & 1, r & 1, g & 1, i & 1) = emu->map_rgbcolor(rr,gg,bb);
#ifdef DEBUG_CRTMON
					emu->out_debugf(_T("palette_64[%02d][%02d] b:%d r:%d g:%d i:%d :%06x")
						,(b >> 1) | ((r >> 1) << 1) | ((g >> 1) << 2) | ((i >> 1) << 3)
						,(b & 1) | ((r & 1) << 1) | ((g & 1) << 2) | ((i & 1) << 3)
						,b,r,g,i,
						COLOR64A(b >> 1, r >> 1, g >> 1, i >> 1, b & 1, r & 1, g & 1, i & 1));
#endif
				}
			}
		}
	}

//	memset(scrntop, 0, sizeof(scrntop));

	screen_left = 0;
	screen_top = 0;
	screen_right = SCREEN_WIDTH + screen_left;
	screen_bottom = SCREEN_HEIGHT + screen_top;
	scrnline1_offset = 0;

	mode_sel = 0;
	ig_mask = true;

	cursor_addr = 0;
	disp_cursor = false;
	for(int i=0; i<32; i++) {
		cur_left[i] = 0;
		cur_right[i] = 0;
	}
	cur_top = 0;
	cur_bottom = 0;

	crtc_chr_clocks = 0;
	crtc_ppch = 16;

	bg_color = 0;
	hireso = 0;
	width_sel = 0;
	page_no = 0;
	gwidth_sel = 0;
	gpage_no = 0;

	top_raster = 0;
	bottom_raster = 0;
	vsync_top_raster = 0;
	vsync_bottom_raster = 0;
	crt_mon_now_raster = 0;

	// synchronizable vertical range
	v_total_min[0] = (LINES_PER_FRAME - 20);
	v_total_max[0] = (LINES_PER_FRAME + 8);
	v_total_min[1] = v_total_min[0] / 2;
	v_total_max[1] = v_total_max[0] / 2;
	v_total_min[2] = v_total_min[0] * 2;
	v_total_max[2] = v_total_max[0] * 2;

//	crt_mon_start_raster = 0;
	for(int i=0; i<VRAM_BUF_HEIGHT; i++) {
		crt_mon_vline[i].left_col = 0;
		crt_mon_vline[i].right_col = 0;
		crt_mon_vline[i].hsync_left_col = 0;
		crt_mon_vline[i].hsync_right_col = 0;
		crt_mon_vline[i].raster = 0;
	}
	left_diff = 0;
	// synchronizable horizontal range
	h_total_min[0] = CHARS_PER_LINE - 3;	// w.40
	h_total_max[0] = CHARS_PER_LINE + 3;
	h_total_min[1] = CHARS_PER_LINE * 2 - 4;	// w.80
	h_total_max[1] = CHARS_PER_LINE * 2 + 4;

	sw = 0;

	//
	memset(crt_mon, 0, sizeof(crt_mon));
	crt_mon_col = 0;
	crt_mon_row = 0;
	crt_mon_start_row = 0;
	crt_mon_l3vaddr = L3_ADDR_VRAM_START;
	crt_mon_s1vtaddr = 0;
	crt_mon_s1vgaddr = 0;
	crt_mon_page = 0;

	crt_mon_cursor_col = CURSOR_MAXPOS;
	crt_mon_cursor_row = CURSOR_MAXPOS;

	disp_page_brg_mask.i = 0;

	skip_frame_count = 0;

	// CRTC 2MHz default
	crtc_chr_clocks = 1;
	crtc_ppch = 8;

	for(int i=0; i<8; i++) {
		ppd[i]=(crtc_ppch == 8) ? 1 : 2;	// pixel per 1dot
		dof[i]=(0x80 >> i);
	}
	// pixel per 1char
//	ppch = (mode_sel & 0x80) ? 8 : 16;

	update_config();

	// register event (unused)
//	register_frame_event(this);
}

void DISPLAY::reset()
{
	load_font_rom_file();

	REG_INTERLACE_SEL = 0;
	REG_IGENABLE = 0;
	ig_mask = true;

	now_cpu_halt = 0;

//	memset(palette_brg, 0, sizeof(palette_brg));

	change_palette();
}

void DISPLAY::update_config()
{
	scanline = config.scan_line;
	afterimage = config.afterimage;
//	keepimage = config.keepimage;
	buf_stepcols = (scanline <= 2 ? 1 : 2);
	crt_mon_stepcols = (scanline <= 2 ? 1 : 2);

	// scanline
	dws_ys = (scanline ? 2 : 1);	// skip step

	// color filter for afterimage
//	memset(scrnline2, 0x00, sizeof(scrnline2));
	memset(scrnflag,  0x00, sizeof(scrnflag));
	if (afterimage == 0) {
		memset(scrnline1, 0x00, sizeof(scrntype) * SCREEN_WIDTH);
		scrnline1_offset = 0;
	} else if (afterimage == 1) {
		scrnline1_offset = SCREEN_WIDTH;

		Rloss = 64 * (scanline <= 1 ? 1 : scanline);
		Gloss = 32 * (scanline <= 1 ? 1 : scanline);
		Bloss = 96 * (scanline <= 1 ? 1 : scanline);
		if (Rloss > 255) Rloss = 255;
		if (Gloss > 255) Gloss = 255;
		if (Bloss > 255) Bloss = 255; 
		Rloss <<= Rshift;
		Gloss <<= Gshift;
		Bloss <<= Bshift;
	} else if (afterimage == 2) {
		scrnline1_offset = SCREEN_WIDTH;
	}

//	if (keepimage > 0) {
//		memset(scrntop, 0, sizeof(scrntop));
//	}
}

void DISPLAY::load_font_rom_file()
{
	const _TCHAR *app_path, *rom_path[2];
	_TCHAR file_path[_MAX_PATH];

	rom_path[0] = config.rom_path;
	rom_path[1] = emu->application_path();

	FILEIO* fio = new FILEIO();
	for(int i=0; i<2; i++) {
		app_path = rom_path[i];

		if (!font_rom_loaded && !l3font_rom_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sFONT.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(font, 0x1000, 1);
				fio->Fclose();
				l3font_rom_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("FONT.ROM was loaded."));
			}
		}
		if (!font_rom_loaded && !l3font_rom_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sS1FONT.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(font, 0x2000, 1);
				fio->Fclose();
				font_rom_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("S1FONT.ROM was loaded."));
			}
		}
		if (font_rom_loaded || l3font_rom_loaded) break;
	}
	delete fio;

	if (!l3font_rom_loaded) {
		if (!font_rom_loaded_at_first) {
			emu->out_log(LOG_WARN, (CT)_("FONT.ROM couldn't be loaded."));
		}
		if (!font_rom_loaded) {
			if (!font_rom_loaded_at_first) {
				emu->out_log(LOG_WARN, (CT)_("S1FONT.ROM couldn't be loaded."));
			}
			memset(chrfont, 0, sizeof(chrfont));
		} else {
			// convert font data to useful array data
			set_font_data();
			conv_to_l3font_data();
			font_rom_loaded = false;
			l3font_rom_loaded = true;
		}
	} else {
		// convert font data to useful array data
		set_l3font_data();
	}
	font_rom_loaded_at_first = true;
}

void DISPLAY::write_io8(uint32 addr, uint32 data)
{
	addr &= 0xffff;
	switch(addr) {
	case 0xfe20:
		// bmsk color graphic color
		REG_BMSK_COLOR = (data & 0x87);
		SIG_BMSK_DW = (BMSK_COLOR_BMSK ? (SIG_BMSK_DW | 2) : (SIG_BMSK_DW & ~2));
		break;
	case 0xfe21:
		// active graphic plane
		REG_ACTIVE_PLANE = (data & 0x0f);
		SIG_BMSK_DW = (ACTIVE_PLANE_DW ? (SIG_BMSK_DW | 1) : (SIG_BMSK_DW & ~1));
		break;
	case 0xfe23:
		// display page number
		REG_DISP_PAGE = (data & 0xff);

		set_disp_page();
		break;
	case 0xfe24:
		// scrn mode
		REG_SCRN_MODE = (data & 0xff);

		if (NOW_S1_MODE) {
			set_s1_screen_mode();
		}
		break;
	case 0xfe25:
	case 0xfe26:
	case 0xfe27:
		// graphic color
		REG_GCOLOR[addr - 0xfe25] = (data & 0xff);
		break;
	case 0xfe28:
	case 0xfe29:
	case 0xfe2a:
	case 0xfe2b:
	case 0xfe2c:
	case 0xfe2d:
	case 0xfe2e:
	case 0xfe2f:
		// palette
		REG_PALETTE[addr - 0xfe28] = (data & 0x0f);
//		set_palette_brg(addr - 0xfe28);

//		emu->out_debugf("dispw a:%04x d:%02x",addr,data);
		break;
	case 0xffd0:
		// mode sel
		mode_sel = (data & 0xff);

		bg_color = mode_sel & 0x0f;				// bg color

		if (NOW_L3_MODE) {
			bg_color |= 0x08;					// bright

			hireso = (mode_sel & 0x40) ? 0 : 1;		// 0:normal 1:hireso
			width_sel = (mode_sel & 0x80) ? 1 : 0;	// 0:w.40  1:w.80

			for(int i=0; i<8; i++) {
				ppd[i]=width_sel ? 1 : 2;	// pixel per 1dot
				dof[i]=(0x80 >> i);
			}
			// pixel per 1char
//			ppch = (mode_sel & 0x80) ? 8 : 16;
		}
		break;
	case 0xffd6:
		// interlace sel
		REG_INTERLACE_SEL = ((data & 0x08) >> 3);	// 0:noninterlace char  1:interlace char
		break;
	case 0xffe9:
		// ig mode reg
		REG_IGENABLE = (data & 0x01);
		ig_mask &= ((data & 0x100) == 0);
//		REG_IGMODE = (ig_mask ? (REG_IGMODE & ~0x80) : (REG_IGMODE | 0x80));
		break;
	}
}

uint32 DISPLAY::read_io8(uint32 addr)
{
	uint32 data = 0xff;
	switch(addr & 0xffff) {
	case 0xfe28:
	case 0xfe29:
	case 0xfe2a:
	case 0xfe2b:
	case 0xfe2c:
	case 0xfe2d:
	case 0xfe2e:
	case 0xfe2f:
		// palette cannot be read
		// when disptmg is on, send HALT signal 
//		if (DISPTMG_ON && !now_cpu_halt) {
		// when hsync and vsync is off, send HALT signal 
		if (!HVSYNC_ON && !now_cpu_halt) {
			d_board->write_signal(SIG_CPU_HALT, SIG_HALT_PALETTE_MASK, SIG_HALT_PALETTE_MASK);
			now_cpu_halt = 1;
//			emu->out_debugf(_T("palette halt fired: %d"),data);
		}
		break;
	}
	return data;
}

void DISPLAY::write_signal(int id, uint32 data, uint32 mask)
{
	switch(id) {
		// Lightpen not supported
#if 0
		case SIG_DISPLAY_LPSTB:
			{
			int width = regs[1];
			int dotx = (data >> 16) & 0xffff;
			int doty = (data & 0xffff);
			int x = 0;
			int y = 0;
			int offset = (width_sel ? 5 : 4);
			uint32 newaddr;
			t_crt_mon_vline *vline = &crt_mon_vline[doty >> 1];

			if (dotx < vline->left_col) { dotx = vline->right_col - 1; doty-=16; }
			if (dotx >= vline->right_col) { dotx = vline->left_col; doty+=16; }
			if (doty < top_raster) doty = top_raster;
			if (doty >= bottom_raster) doty = bottom_raster - 1;

			x = (dotx - vline->left_col) >> (width_sel ? 3 : 4);
			y = (doty - top_raster) >> 4;
			newaddr = ((regs[12] << 8)+regs[13]) + y * width + x + offset;
			regs[16] = (newaddr >> 8) & 0x3f;
			regs[17] = newaddr & 0xff;

//			emu->out_debugf("lp dx:%03d dy:%03d x:%02d y:%02d addr:%02x%02x"
//				,dotx,doty,x,y,regs[16],regs[17]);
			}
			break;
#endif
		case SIG_DISPLAY_DISPTMG:
			// display area
			REG_IGMODE = ((data & mask) ? (REG_IGMODE | DISPTMG_MASK) : (REG_IGMODE & ~DISPTMG_MASK));
//			emu->out_debugf(_T("disptmg %06lld IGM:%02x  %d:%d"), current_clock() & 0xffffff, REG_IGMODE, data, mask);
			break;
		case SIG_DISPLAY_HSYNC:
			// hsync
			REG_IGMODE = ((data & mask) ? (REG_IGMODE | HSYNC_MASK) : (REG_IGMODE & ~HSYNC_MASK));
			// release halt signal
			if (now_cpu_halt) {
				now_cpu_halt = 0;
				d_board->write_signal(SIG_CPU_HALT, 0, SIG_HALT_PALETTE_MASK);
//				emu->out_debugf(_T("hsync halt released: %d %d"),data,mask);
			}
			// TODO: PAE flag is different from real machine
			REG_IGMODE = (HVSYNC_ON ? REG_IGMODE | PAE_MASK : REG_IGMODE & ~PAE_MASK);

//			emu->out_debugf(_T("hsync   %06lld IGM:%02x  %d:%d"), current_clock() & 0xffffff, REG_IGMODE, data, mask);
			break;
		case SIG_DISPLAY_VSYNC:
			if ((data & mask) == 0) {
				REG_IGMODE = (REG_IGMODE & ~VSYNC_MASK);
				// synchronize display vertical raster
				for(int i=0; i<3; i++) {
					if (v_total_min[i] <= (*crtc_vt_total) && (*crtc_vt_total) <= v_total_max[i]) {
						crt_mon_now_raster = 0;
						break;
					}
				}

#ifdef DEBUG_CRTMON
				emu->out_debugf(_T("DISPLAY:SIG_DISPLAY_VSYNC_OFF: crtc:%d nras:%d cmr:%d cmc:%d cmras:%d top:%d btm:%d vtop:%d vbtm:%d")
					,(*crtc_vt_count), crt_mon_now_raster
					,crt_mon_row, crt_mon_col, (*crtc_ra)
					,top_raster, bottom_raster, vsync_top_raster, vsync_bottom_raster
				);
#endif
			} else {
				REG_IGMODE = (REG_IGMODE | VSYNC_MASK);
			}
			// release halt signal
			if (now_cpu_halt) {
				now_cpu_halt = 0;
				d_board->write_signal(SIG_CPU_HALT, 0, SIG_HALT_PALETTE_MASK);
//				emu->out_debugf(_T("vsync halt released: %d %d"),data,mask);
			}
			// TODO: PAE flag is different from real machine
			REG_IGMODE = (HVSYNC_ON ? REG_IGMODE | PAE_MASK : REG_IGMODE & ~PAE_MASK);

//			emu->out_debugf(_T("vsync   %06lld IGM:%02x  %d:%d ras:%d"), current_clock() & 0xffffff, REG_IGMODE, data, mask, now_raster);
			break;
		case SIG_DISPLAY_CHR_CLOCKS:
			crtc_chr_clocks = (data == CLOCKS_1MHZ ? 0 : 1);
			crtc_ppch = (crtc_chr_clocks ? 8 : 16);
			break;
		case SIG_CPU_RESET:
			// halt on while res signal is on.
			if (!now_reset && (data & mask) && now_cpu_halt == 0) {
				// turn on
				d_board->write_signal(SIG_CPU_HALT, SIG_HALT_PALETTE_MASK, SIG_HALT_PALETTE_MASK);
				now_cpu_halt = 1;
			} else if (now_reset && (data & mask) == 0 && now_cpu_halt) {
				// turn off
				d_board->write_signal(SIG_CPU_HALT, 0, SIG_HALT_PALETTE_MASK);
				now_cpu_halt = 0;
			}
			now_reset = (data & mask) ? true : false;
			if (!now_reset) {
				change_palette();
			}
			break;
	}
}

void DISPLAY::update_cursor()
{
	int cur_left_0;
	int cur_width;
	uint8 cur_st_l;
	uint8 cur_ed_l;

	if(crt_mon_cursor_col >= CURSOR_MAXPOS || crt_mon_cursor_row >= CURSOR_MAXPOS) {
		disp_cursor = false;
	} else {
		disp_cursor = ((*crtc_curdisp) != 0);
	}

	if (disp_cursor) {
		if ((*videomode) == 1) {
			cur_st_l = (regs[10] & 0x0f);
			cur_ed_l = (regs[11] & 0x0f);
		} else {
			cur_st_l = ((regs[10] & 0x0f) << 1);
			cur_ed_l = ((regs[11] & 0x0f) << 1) + 1;
		}

		// cursor draw area
		cur_top = top_raster + (crt_mon_cursor_row << 1) - (crt_mon_vline[crt_mon_cursor_row].raster & 0x7e);
		cur_bottom = cur_top + cur_ed_l;
		cur_top += cur_st_l;

		cur_left_0 = crt_mon_vline[crt_mon_cursor_row].left_col + (crt_mon_cursor_col + (*curdisp_skew) + (NOW_L3_MODE ? config.curdisp_skew : 0) - 1 - (*disptmg_skew) + config.disptmg_skew) * crtc_ppch;
		cur_width = crtc_ppch + ((NOW_S1_MODE && width_sel == 0) ? crtc_ppch : 0);
		for(int i=0; i < ((*crtc_max_ra) << (1-(*videomode))); i++) {
			cur_left[i] = cur_left_0;
			cur_right[i] = cur_left[i] + cur_width;
		}
	}
	crt_mon_cursor_col = CURSOR_MAXPOS;
	crt_mon_cursor_row = CURSOR_MAXPOS;
}

void DISPLAY::set_display_size(int left, int top, int right, int bottom)
{
	screen_left = left;
	screen_right = right;
	screen_top = top;
	screen_bottom = bottom;
}

void DISPLAY::draw_screen()
{
	if (emu->now_skip_frame()) {
		skip_frame_count++;
	} else {
		skip_frame_count = 0;
	}
	if ((afterimage == 0 && skip_frame_count >= 1)
	||  (afterimage > 0 && skip_frame_count >= 2)) {
		// clear cursor
//		disp_cursor = false;
//		crt_mon_cursor_col = CURSOR_MAXPOS;
//		crt_mon_cursor_row = CURSOR_MAXPOS;

		return;
	}

	// update cursor
	update_cursor();

	// render screen
	if (afterimage == 0) draw_screen_sub();
	else if (afterimage == 1) draw_screen_sub_afterimage1();
	else draw_screen_sub_afterimage2();

	if (!(*vm_pause) || scanline >= 2) sw = ((sw + 1) & 3);
	if (scanline <= 1 && *crtc_odd_line) sw = *crtc_odd_line;
}

uint8 DISPLAY::get_font_data(int interlace, int chr_code, int y)
{
	int code = 0;
	int offset = 0;
	if (interlace == 1) {
		// interlace mode char
		code = (chr_code | 0x100) << 4;	// 16byte per 1char
		offset = (y & 0xf);	// 0 - 15
	} else {
		// non interlace mode char
		code = chr_code << 4;	// 16byte per 1char
		offset = (y & 0x7);	// 0 - 7
	}
	return font[code + offset];
}

void DISPLAY::set_font_data()
{
	for(int i = 0; i < 2; i++) {
		for(int c = 0; c < 256; c++) {
			for(int l = 0; l < 16; l++) {
				chrfont[i][c][l] = get_font_data(i, c, l);
			}
		}
	}
}

uint8 DISPLAY::get_l3font_data(int interlace, int chr_code, int y)
{
	int code = 0;
	int offset = 0;
	if (interlace == 1) {
		// interlace mode char
		if (chr_code < 0x80) {
			code = chr_code << 4;	// 16byte per 1char
			offset = (y & 0x0e);	// 0 0 2 2 4 4 6 6 8 8 a a c c e e
		} else {
			code = chr_code << 4;	// 16byte per 1char
			offset = (y & 0x0f);	// 0 1 2 3 4 5 6 7 8 9 a b c d e f
		}
	} else {
		// non interlace mode char
		if (chr_code < 0x80) {
			// ascii code
			code = chr_code << 4;	// 16byte per 1char
		} else {
			code = ((chr_code - 0x80) << 4) + 1;
		}
		offset = ((y << 1) & 0x0e);	// 0 2 4 6 8 a c e 0 2 4 6 8 a c e
	}
	return font[code + offset];
}

void DISPLAY::set_l3font_data()
{
	for(int i = 0; i < 2; i++) {
		for(int c = 0; c < 256; c++) {
			for(int l = 0; l < 16; l++) {
				chrfont[i][c][l] = get_l3font_data(i, c, l);
			}
		}
	}
}

void DISPLAY::conv_to_l3font_data()
{
	int code = 0;
	int offset = 0;
	memset(font, 0, sizeof(font));
	for(int chr_code = 0; chr_code < 256; chr_code++) {
		for(int y=0; y<8; y++) {
			// non interlace mode char
			if (chr_code < 0x80) {
				// ascii code
				code = chr_code << 4;	// 16byte per 1char
			} else {
				code = ((chr_code - 0x80) << 4) + 1;
			}
			offset = ((y << 1) & 0x0e);	// 0 2 4 6 8 a c e 0 2 4 6 8 a c e
			font[code + offset] = chrfont[0][chr_code][y];
		}
	}
	for(int chr_code = 128; chr_code < 256; chr_code++) {
		for(int y=0; y<16; y++) {
			// interlace mode char
			code = chr_code << 4;	// 16byte per 1char
			offset = (y & 0x0f);	// 0 1 2 3 4 5 6 7 8 9 a b c d e f
			font[code + offset] = chrfont[1][chr_code][y];
		}
	}
}

#define COLOR_REVERSE(fg,bg) { \
	uint8 c; \
	c = (bg); \
	(bg) = (fg); \
	(fg) = c; \
}

#define DECODE_TO_PALETTE16_NUM(code, outc) { \
	outc.b.b = (code & 1) ? 1 : 0; \
	outc.b.r = (code & 2) ? 1 : 0; \
	outc.b.g = (code & 4) ? 1 : 0; \
	outc.b.i = (code & 8) ? 1 : 0; \
}

#define DECODE_TO_PALETTE64_NUM(code, outc) { \
	outc.b.b = (code & 1) ? ((code & 8) ? 3 : 2) : 0; \
	outc.b.r = (code & 2) ? ((code & 8) ? 3 : 2) : 0; \
	outc.b.g = (code & 4) ? ((code & 8) ? 3 : 2) : 0; \
}

#define SET_SCRNLINE2M(ii) dws_scrnline2m = ((ii) ? 0 : 0xffffffff) 
#define SET_SCRN_DOT(xx,dd) dws_scrn[xx] = (dd | (dws_scrnline1p[xx] & dws_scrnline2m))
#define SET_SCRN_DOT_COLOR16(xx,bb,rr,gg,ii) dws_scrn[xx] = (COLOR16A(bb,rr,gg,ii) | (dws_scrnline1p[xx] & dws_scrnline2m))
#define SET_SCRN_FILTER(xx) dws_scrnf[xx] &= dws_scrnline2m

#define DISP_OUT_OF_DISPTMG(p_vline,xx,pp,bp) { \
	if (p_vline->hsync_left_col <= xx && xx < p_vline->hsync_right_col) { \
		pp.i = 0; \
	} else { \
		pp = bp; \
	} \
	SET_SCRN_DOT_COLOR16(xx, pp.b.b, pp.b.r, pp.b.g, pp.b.i); \
	SET_SCRN_FILTER(xx); \
}

#define DISP_TEXT_CHAR_ONE_DOT(pos, code) { \
	dws_p_p = (code & dof[pos]) ? dws_t_p : dws_b_p; \
	dws_dot = COLOR16A(dws_p_p.b.b,dws_p_p.b.r,dws_p_p.b.g,dws_p_p.b.i); \
	SET_SCRNLINE2M(dws_p_p.i & BRG_COLOR_IMASK); \
	dws_dtm = dws_x+dws_bi+ppd[pos]; \
	if (dws_dtm > SCREEN_WIDTH) dws_dtm = SCREEN_WIDTH; \
	for(dws_dt = dws_x+dws_bi; dws_dt < dws_dtm; dws_dt++) { \
		SET_SCRN_DOT(dws_dt,dws_dot); \
		SET_SCRN_FILTER(dws_dt); \
	} \
}

#define DISP_TEXT_CHAR_MAIN(code) { \
	dws_bi = 0; \
	for(dws_bt = 0; dws_bt < 8; dws_bt++) { \
		DISP_TEXT_CHAR_ONE_DOT(dws_bt, code); \
		dws_bi += ppd[dws_bt]; \
	} \
}

#define DISP_L3_TEXT_CHAR(code) { \
	dws_pattern = chrfont[REG_INTERLACE_SEL][code][(dws_l >> (1 - (*videomode))) & 0xf]; \
\
	if (dws_disp_cursor_char) dws_pattern = ~dws_pattern; \
	if (dws_text_cram & 8) dws_pattern = ~dws_pattern; \
	DISP_TEXT_CHAR_MAIN(dws_pattern) \
}

#define DISP_IG_PATTERN_ONE_DOT(ci, bt) { \
	if (FLG_ORIG_CURIG && dws_disp_cursor_char) { \
		dws_p_p = dws_t_p; \
	} else { \
		dws_pattern = ((ci[0].b.g & dof[bt]) ? 4 : 0) | ((ci[0].b.r & dof[bt]) ? 2 : 0) | ((ci[0].b.b & dof[bt]) ? 1 : 0); \
		DECODE_TO_PALETTE16_NUM(dws_pattern, dws_p_p); \
		if (dws_p_p.i == 0) dws_p_p = dws_b_p; \
		dws_p_p.b.i = 1; \
	} \
	dws_dot = COLOR16A(dws_p_p.b.b,dws_p_p.b.r,dws_p_p.b.g,dws_p_p.b.i); \
	SET_SCRNLINE2M(dws_p_p.i & BRG_COLOR_IMASK); \
	dws_dtm = dws_x+dws_bi+ppd[bt]; \
	if (dws_dtm > SCREEN_WIDTH) dws_dtm = SCREEN_WIDTH; \
	for(dws_dt = dws_x+dws_bi; dws_dt < dws_dtm; dws_dt++) { \
		SET_SCRN_DOT(dws_dt,dws_dot); \
		SET_SCRN_FILTER(dws_dt); \
	} \
}

#define DISP_IG_PATTERN_MAIN(ci) { \
	dws_bi = 0; \
	for(dws_bt = 0; dws_bt < 8; dws_bt++) { \
		DISP_IG_PATTERN_ONE_DOT(ci, dws_bt) \
		dws_bi += ppd[dws_bt]; \
	} \
}

#define DISP_L3_IG_PATTERN(code) { \
	dws_ll = dws_l >> dws_int_vid; \
	dws_ct[0].b.b = ig_ram[code*8+dws_ll]; \
	dws_ct[0].b.r = ig_ram[code*8+dws_ll+IGRAM_SIZE]; \
	dws_ct[0].b.g = ig_ram[code*8+dws_ll+IGRAM_SIZE+IGRAM_SIZE]; \
	if (FLG_ORIG_CURIG && (dws_text_cram & 8)) dws_ct[0].i = ~dws_ct[0].i; \
\
	DISP_IG_PATTERN_MAIN(dws_ct) \
}

#define DISP_WHITE_SPACE { \
	dws_c_f = 15; \
	DECODE_TO_PALETTE16_NUM(dws_c_f, dws_p_p); \
	dws_dot = COLOR16A(dws_p_p.b.b,dws_p_p.b.r,dws_p_p.b.g,dws_p_p.b.i); \
	SET_SCRNLINE2M(dws_p_p.i & BRG_COLOR_IMASK); \
	dws_dtm = dws_x+ppch; \
	if (dws_dtm > SCREEN_WIDTH) dws_dtm = SCREEN_WIDTH; \
	for(dws_dt = dws_x; dws_dt < dws_dtm; dws_dt++) { \
		SET_SCRN_DOT(dws_dt,dws_dot); \
		SET_SCRN_FILTER(dws_dt); \
	} \
}

#define	DISP_L3_GRAPHIC_HIRESO_MAIN(code) { \
	dws_bi = 0; \
	for(dws_bt = 0; dws_bt < 8; dws_bt++) { \
		dws_p_p = (code & dof[dws_bt]) ? dws_t_p : dws_b_p; \
		dws_dot = COLOR16A(dws_p_p.b.b,dws_p_p.b.r,dws_p_p.b.g,dws_p_p.b.i); \
		SET_SCRNLINE2M(dws_p_p.i & BRG_COLOR_IMASK); \
		dws_dtm = dws_x+dws_bi+ppd[dws_bt]; \
		if (dws_dtm > SCREEN_WIDTH) dws_dtm = SCREEN_WIDTH; \
		for(dws_dt = dws_x+dws_bi; dws_dt < dws_dtm; dws_dt++) { \
			SET_SCRN_DOT(dws_dt,dws_dot); \
			SET_SCRN_FILTER(dws_dt); \
		} \
		dws_bi += ppd[dws_bt]; \
	} \
}

#define	DISP_L3_GRAPHIC_HIRESO(code) \
	DISP_L3_GRAPHIC_HIRESO_MAIN(code)


#define DISP_L3_GRAPHIC_NORMAL_MAIN(code) { \
	dws_bi = 0; \
	for (dws_bt = 0; dws_bt < 8; dws_bt += 4) { \
		dws_p_p = (code & ((0x10 >> dws_bt) << ((dws_l >> (1 + dws_int_vid)) & 3))) ? dws_t_p : dws_b_p; \
		dws_dot = COLOR16A(dws_p_p.b.b,dws_p_p.b.r,dws_p_p.b.g,dws_p_p.b.i); \
		SET_SCRNLINE2M(dws_p_p.i & BRG_COLOR_IMASK); \
		dws_dtm = dws_x+dws_bi+ppd[dws_bt]+ppd[dws_bt+1]+ppd[dws_bt+2]+ppd[dws_bt+3]; \
		if (dws_dtm > SCREEN_WIDTH) dws_dtm = SCREEN_WIDTH; \
		for(dws_dt = dws_x+dws_bi; dws_dt < dws_dtm; dws_dt++) { \
			SET_SCRN_DOT(dws_dt,dws_dot); \
			SET_SCRN_FILTER(dws_dt); \
		} \
		dws_bi += ppd[dws_bt] + ppd[dws_bt + 1] + ppd[dws_bt + 2] + ppd[dws_bt + 3]; \
	} \
}

#define DISP_L3_GRAPHIC_NORMAL(code) \
	DISP_L3_GRAPHIC_NORMAL_MAIN(code)

#define EXPAND_BRG_1BIT(code, in8, outc) { \
	if (code & 1) { outc.b.b |= in8; } \
	if (code & 2) { outc.b.r |= in8; } \
	if (code & 4) { outc.b.g |= in8; } \
	if (code & 8) { outc.b.i |= in8; } \
}

#define EXPAND_BRG_2BIT(code, in8, outc) { \
	if (code & 1) { outc[0].b.b |= in8; } \
	if (code & 2) { outc[0].b.r |= in8; } \
	if (code & 4) { outc[0].b.g |= in8; } \
	if (code & 8) { outc[0].b.i |= in8; } \
	outc[1].i = outc[0].i; \
}

#define EXPAND_BRG_2BIT_I(code, in8, outc) { \
	wk1 = in8; wk2 = (code & 8); \
	if (code & 1) { outc[0].b.b |= wk1; outc[1].b.b |= (wk2 ? wk1 : 0); } \
	if (code & 2) { outc[0].b.r |= wk1; outc[1].b.r |= (wk2 ? wk1 : 0); } \
	if (code & 4) { outc[0].b.g |= wk1; outc[1].b.g |= (wk2 ? wk1 : 0); } \
}

#define INIT_BRG_1(data,out) \
	out.i = data;

#define INIT_BRG_2(data,out) \
	out[0].i = out[1].i = data;

#define ENCODE_COLOR_ONE_DOT_1(pos, inc, outc) { \
	iwk1.i = ((inc.i & (0x01010101 << pos)) >> pos); \
	outc[pos] = iwk1.b.b | (iwk1.b.r << 1) | (iwk1.b.g << 2); \
}

#define ENCODE_PALETTE_ONE_DOT_1(p_vline, pos, inc, outc) { \
	iwk1.i = ((inc.i & (0x01010101 << pos)) >> pos); \
	outc[pos] = p_vline->pal[iwk1.b.b | (iwk1.b.r << 1) | (iwk1.b.g << 2)]; \
}

#define ENCODE_COLOR_ONE_DOT_2(pos, inc, outc) { \
	iwk1.i = ((inc[0].i & (0x01010101 << pos)) >> pos); \
	outc[0][pos] = iwk1.b.b | (iwk1.b.r << 1) | (iwk1.b.g << 2); \
	iwk1.i = ((inc[1].i & (0x01010101 << pos)) >> pos); \
	outc[1][pos] = iwk1.b.b | (iwk1.b.r << 1) | (iwk1.b.g << 2); \
}

#define ENCODE_PALETTE_ONE_DOT_2(p_vline, pos, inc, outc) { \
	iwk1.i = ((inc[0].i & (0x01010101 << pos)) >> pos); \
	outc[0][pos] = p_vline->pal[iwk1.b.b | (iwk1.b.r << 1) | (iwk1.b.g << 2)]; \
	iwk1.i = ((inc[1].i & (0x01010101 << pos)) >> pos); \
	outc[1][pos] = p_vline->pal[iwk1.b.b | (iwk1.b.r << 1) | (iwk1.b.g << 2)]; \
}

#define DECODE_BRG_2BIT_I(pos, inc, outc) { \
	iwk2.i = (inc.i << pos); \
	if (inc.i == 0x1000000) { \
		outc[1].i |= (0x010101 << pos); \
	} else { \
		outc[0].i |= iwk2.i; \
		if (inc.b.i) outc[1].i |= iwk2.i; \
	} \
}

#define CREATE_MASK_1(inc,outc) { \
	wk1u8 = ~(inc.b.b | inc.b.r | inc.b.g | inc.b.i); \
	outc.b.b = outc.b.r = outc.b.g = outc.b.i = wk1u8; \
}

#define CREATE_MASK_2(inc,outc) { \
	wk1u8 = ~(inc[0].b.b | inc[0].b.r | inc[0].b.g | inc[0].b.i | inc[1].b.b | inc[1].b.r | inc[1].b.g | inc[1].b.i); \
	outc.b.r = outc.b.g = outc.b.b = outc.b.i = wk1u8; \
}

#define COPY_BRG_1(inc,outc) { \
	outc.i = inc.i; \
}

#define COPY_BRG_2(inc,outc) { \
	outc[0].i = inc[0].i; outc[1].i = inc[1].i; \
}

#define COPY_BRG_1_WITH_MASK(inc,outc,msk) { \
	outc.i = (inc.i & msk.i); \
}

#define COPY_BRG_2_WITH_MASK(inc,outc,msk) { \
	outc[0].i = (inc[0].i & msk.i); outc[1].i = (inc[1].i & msk.i); \
}

#define ADD_BRG_1(inc,outc) { \
	outc.i |= inc.i; \
}

#define ADD_BRG_2(inc,outc) { \
	outc[0].i |= inc[0].i; outc[1].i |= inc[1].i; \
}

#define ADD_BRG_1_WITH_MASK(inc,outc,msk) { \
	outc.i |= (inc.i & msk.i); \
}

#define ADD_BRG_2_WITH_MASK(inc,outc,msk) { \
	outc[0].i |= (inc[0].i & msk.i); outc[1].i |= (inc[1].i & msk.i); \
}

#define REVERSE_BRG_1(outc) { \
	outc.i = ~outc.i; \
}

#define REVERSE_BRG_2(outc) { \
	outc[0].i = ~outc[0].i;	outc[1].i = ~outc[1].i; \
}

#define DECODE_BRG_1BIT_TO_PALETTE16(pos,inc,outc) { \
	iwk1.i = (0x80808080 >> pos); iwk2.i = (7 - pos); \
	outc.i = ((inc[0].i & iwk1.i) >> iwk2.i); \
}

#define DECODE_BRG_2BIT_TO_PALETTE64_R(pos,inc,outc) { \
	iwk1.i = (0x80808080 >> pos); iwk2.i = (7 - pos); \
	outc.i = (((inc[0].i & iwk1.i) >> iwk2.i) | (((inc[1].i & iwk1.i) >> iwk2.i) << 1)); \
}

#define DECODE_BRG_2BIT_TO_PALETTE64(pos,inc,outc) { \
	iwk1.i = (0x80808080 >> pos); iwk2.i = (7 - pos); \
	outc.i = ((((inc[0].i & iwk1.i) >> iwk2.i) << 1) | ((inc[1].i & iwk1.i) >> iwk2.i)); \
}

#define EXPAND_UINT8_LEFT(inp,outp) { \
	outp = ((inp) & 0x01) | (((inp) & 0x02) << 1) | (((inp) & 0x04) << 2) | (((inp) & 0x08) << 3); \
	outp |= ((outp) << 1); \
}							
#define EXPAND_UINT8_RIGHT(inp,outp) { \
	outp = (((inp) & 0x10) >> 4) | (((inp) & 0x20) >> 3) | (((inp) & 0x40) >> 2) | (((inp) & 0x80) >> 1); \
	outp |= ((outp) << 1); \
}

#define EXPAND_UINT32_LEFT(inp,outp) { \
	outp = ((inp) & 0x01010101) | (((inp) & 0x02020202) << 1) | (((inp) & 0x04040404) << 2) | (((inp) & 0x08080808) << 3); \
	outp |= ((outp) << 1); \
}							
#define EXPAND_UINT32_RIGHT(inp,outp) { \
	outp = (((inp) & 0x10101010) >> 4) | (((inp) & 0x20202020) >> 3) | (((inp) & 0x40404040) >> 2) | (((inp) & 0x80808080) >> 1); \
	outp |= ((outp) << 1); \
}

#define SET_YAXIS_POSITION { \
	if ((*disptmg_skew) == 3) { \
		dws_y0[0] = (screen_top | dws_yd); dws_y1[0] = screen_bottom;	/* no drawing */ \
		dws_y0[1] = 0; dws_y1[1] = 0; \
		dws_y0[2] = 0; dws_y1[2] = 0; \
		dws_ii[0] = 1; dws_ii[1] = 1; \
	} \
	else if (top_raster <  bottom_raster) { \
		dws_y0[0] = (screen_top | dws_yd); dws_y1[0] = (screen_bottom < top_raster ? screen_bottom : top_raster);	/* top padding */ \
		dws_y0[1] = (bottom_raster | dws_yd); dws_y1[1] = screen_bottom;	/* bottom padding */ \
		dws_y0[2] = ((top_raster > screen_top ? top_raster : screen_top) | dws_yd); dws_y1[2] = (screen_bottom < bottom_raster ? screen_bottom : bottom_raster); \
		dws_ii[0] = 2; dws_ii[1] = 3; \
	} \
	else if (top_raster >= bottom_raster) { \
		dws_y0[0] = ((bottom_raster > screen_top ? bottom_raster : screen_top) | dws_yd); dws_y1[0] = (screen_bottom < top_raster ? screen_bottom : top_raster); \
		dws_y0[1] = (screen_top | dws_yd); dws_y1[1] = (screen_bottom < bottom_raster ? screen_bottom : bottom_raster);	/* top drawing */ \
		dws_y0[2] = (top_raster | dws_yd); dws_y1[2] = screen_bottom;	/* bottom drawing */ \
		dws_ii[0] = 1; dws_ii[1] = 3; \
	} \
	else { \
		dws_y0[0] = (screen_top | dws_yd); dws_y1[1] = screen_bottom; \
		dws_y0[1] = 0; dws_y1[1] = 0; \
		dws_y0[2] = 0; dws_y1[2] = 0; \
		dws_ii[0] = 0; dws_ii[1] = 1; \
	} \
}

#undef SET_SCRNLINE2M
#undef SET_SCRN_DOT
#undef SET_SCRN_DOT_COLOR16
#undef SET_SCRN_FILTER

#define INIT_SCRN_DOT(yy) dws_scrn = dws_scrn0 + dws_scrn_offset * (yy)
#define INIT_SCRN_FILTER(yy)
#define INIT_SCRNLINE1(yy)
#define SET_SCRNLINE2M(ii)
#define SET_SCRN_DOT(xx,dd) dws_scrn[xx] = (dd)
#define SET_SCRN_DOT_COLOR16(xx,bb,rr,gg,ii) dws_scrn[xx] = COLOR16A(bb,rr,gg,ii)
#define SET_SCRN_FILTER(xx)
#define ADD_SCRN_DOT dws_scrn += dws_scrn_offset * dws_ys
#define ADD_SCRN_FILTER
#define ADD_SCRNLINE1

void DISPLAY::draw_screen_sub()
{
	dws_raster_per_char = ((*crtc_max_ra) << (1 - (*videomode)));
	dws_disp_cursor_line = false;
	dws_disp_cursor_char = false;
	dws_scrn_offset = emu->screen_buffer_offset();
	dws_scrn0 = emu->screen_buffer(0);
	dws_dot = 0;
	dws_buf_x = 0;

#include "displaysub.h"
}

#undef INIT_SCRN_FILTER
#undef INIT_SCRNLINE1
#undef SET_SCRNLINE2M
#undef SET_SCRN_DOT
#undef SET_SCRN_DOT_COLOR16
#undef SET_SCRN_FILTER
#undef ADD_SCRNLINE1
#undef ADD_SCRN_FILTER

#define SCRNTYPE_SELDOT(newdot, olddot) \
	(((newdot & Rmask) > (olddot & Rmask) ? (newdot & Rmask) : (olddot & Rmask)) \
	| ((newdot & Gmask) > (olddot & Gmask) ? (newdot & Gmask) : (olddot & Gmask)) \
	| ((newdot & Bmask) > (olddot & Bmask) ? (newdot & Bmask) : (olddot & Bmask)) \
	| Amask)

#define INIT_SCRNLINE1(yy)   dws_scrnline1p = scrnline1[0] + scrnline1_offset * (yy)
#define INIT_SCRN_FILTER(yy)
#define SET_SCRNLINE2M(ii)
#define SET_SCRN_DOT(xx,dd) dws_scrn[xx] = SCRNTYPE_SELDOT(dd, dws_scrnline1p[xx])
#define SET_SCRN_DOT_COLOR16(xx,bb,rr,gg,ii) dws_scrn[xx] = SCRNTYPE_SELDOT(COLOR16A(bb,rr,gg,ii), dws_scrnline1p[xx])
#define SET_SCRN_FILTER(xx)
#define ADD_SCRNLINE1   dws_scrnline1p += scrnline1_offset * dws_ys
#define ADD_SCRN_FILTER

void DISPLAY::draw_screen_sub_afterimage1()
{
	dws_raster_per_char = ((*crtc_max_ra) << (1 - (*videomode)));
	dws_disp_cursor_line = false;
	dws_disp_cursor_char = false;
	dws_scrn_offset = emu->screen_buffer_offset();
	dws_scrn0 = emu->screen_buffer(0);
	dws_dot = 0;
	dws_buf_x = 0;

	// color filter for afterimage
	dws_scrn = dws_scrn0 + dws_scrn_offset * screen_top;
	for(dws_y = screen_top; dws_y < screen_bottom; dws_y++) {
		for(dws_x = screen_left; dws_x < screen_right; dws_x++) {
			dws_dot = dws_scrn[dws_x];
			dws_r = (dws_dot & Rmask);
			dws_g = (dws_dot & Gmask);
			dws_b = (dws_dot & Bmask);

			if (dws_r > Rloss) dws_r = dws_r - Rloss;
			else dws_r = 0;
			if (dws_g > Gloss) dws_g = dws_g - Gloss;
			else dws_g = 0;
			if (dws_b > Bloss) dws_b = dws_b - Bloss;
			else dws_b = 0;

			dws_dot = dws_r | dws_g | dws_b;

			scrnline1[dws_y][dws_x] = dws_dot;
		}
		dws_scrn += dws_scrn_offset;
	}

#include "displaysub.h"
}

#undef INIT_SCRN_FILTER
#undef INIT_SCRNLINE1
#undef SET_SCRNLINE2M
#undef SET_SCRN_DOT
#undef SET_SCRN_DOT_COLOR16
#undef SET_SCRN_FILTER
#undef ADD_SCRNLINE1
#undef ADD_SCRN_FILTER

#define INIT_SCRNLINE1(yy)   dws_scrnline1p = scrnline1[0] + scrnline1_offset * (yy)
#define INIT_SCRN_FILTER(yy) dws_scrnf = scrnflag[0] + dws_scrn_offset * (yy)
#define SET_SCRNLINE2M(ii) dws_scrnline2m = ((ii) ? 0 : 0xffffffff) 
#define SET_SCRN_DOT(xx,dd) dws_scrn[xx] = (dd | (dws_scrnline1p[xx] & dws_scrnline2m))
#define SET_SCRN_DOT_COLOR16(xx,bb,rr,gg,ii) dws_scrn[xx] = (COLOR16A(bb,rr,gg,ii) | (dws_scrnline1p[xx] & dws_scrnline2m))
#define SET_SCRN_FILTER(xx) dws_scrnf[xx] &= dws_scrnline2m
#define ADD_SCRNLINE1   dws_scrnline1p += scrnline1_offset * dws_ys
#define ADD_SCRN_FILTER dws_scrnf += dws_scrn_offset * dws_ys

void DISPLAY::draw_screen_sub_afterimage2()
{
	dws_raster_per_char = ((*crtc_max_ra) << (1 - (*videomode)));
	dws_disp_cursor_line = false;
	dws_disp_cursor_char = false;
	dws_scrn_offset = emu->screen_buffer_offset();
	dws_scrn0 = emu->screen_buffer(0);
	dws_dot = 0;
	dws_buf_x = 0;

	// color filter for afterimage2
	dws_scrnline2m = 0;
	dws_scrn = dws_scrn0 + dws_scrn_offset * screen_top;
	for(dws_y = screen_top; dws_y < screen_bottom; dws_y++) {
		for(dws_x = screen_left; dws_x < screen_right; dws_x++) {
			dws_dot = dws_scrn[dws_x];

			if (scrnflag[dws_y][dws_x] >= 2) {
				scrnflag[dws_y][dws_x] = 0;
				dws_dot = 0;
			} else {
				scrnflag[dws_y][dws_x]++;
			}
				
			scrnline1[dws_y][dws_x] = dws_dot;
		}
		dws_scrn += dws_scrn_offset;
	}

#include "displaysub.h"
}

#define SET_CRT_MON(row, name, value) \
	crt_mon[row][crt_mon_col].name = (value);

#define ADD_CRT_MON(row, name, value) \
	crt_mon[row][crt_mon_col].name |= (value);

#define GET_CRT_MON(row, name) \
	crt_mon[row][crt_mon_col].name

#define P_SET_CRT_MON(ptr, name, value) \
	ptr->name = (value);

#define P_ADD_CRT_MON(ptr, name, value) \
	ptr->name |= (value);

#define P_GET_CRT_MON(ptr, name) \
	ptr->name

void DISPLAY::update_vram(int v, int clock)
{
	static int left_col;
	static int addr;
	static int addr_offset;
	static uint8 text,color,val,val2;
	static int vtaddr;
	static int vtadiv;
	static int vgaddr;
	static int vgaddr2;
	static int vgadiv;
	static bool now_skip_frame;

	if ((*crtc_vt_count) != crt_mon_row) {
		crt_mon_row = (*crtc_vt_count) % VRAM_BUF_HEIGHT;
		crt_mon_col = (crt_mon_stepcols - 1) * ((crt_mon_row + sw) & 1);
		int row_zero = (*crtc_vt_count) % LINES_PER_FRAME;

		if (row_zero == 0) {
			// disptmg on

			// pixel of height
//			crt_mon_start_raster = ((*crtc_ra) - (*crtc_odd_line)) << (1 - (*videomode));

			if (*crtc_vt_count) {
				// no sync with vsync
				crt_mon_start_row = crt_mon_row;
				top_raster = 0;
				bottom_raster = top_raster + (regs[6] * (*crtc_max_ra) << (1 - (*videomode))) - (*crtc_vt_count) * 2;

				vsync_top_raster = top_raster + (regs[7] * (*crtc_max_ra) << (1 - (*videomode))) - (*crtc_vt_count) * 2;
				vsync_bottom_raster = vsync_top_raster + (((regs[3] & 0xf0) ? (regs[3] >> 4) : 16) << 1);
			} else {
				// if disptmg turn on then top_raster is crt_mon_now_raster.
				crt_mon_start_row = 0;
				top_raster = crt_mon_now_raster * 2;
				top_raster += (SCREEN_HEIGHT - 488) / 2;
				bottom_raster = top_raster + (regs[6] * (*crtc_max_ra) << (1 - (*videomode)));

				vsync_top_raster = top_raster + (regs[7] * (*crtc_max_ra) << (1 - (*videomode)));
				vsync_bottom_raster = vsync_top_raster + (((regs[3] & 0xf0) ? (regs[3] >> 4) : 16) << 1);
			}

			// synchronize horizontal total
			if (h_total_min[crtc_chr_clocks] <= regs[0] && regs[0] <= h_total_max[crtc_chr_clocks]) {
				left_diff = 0;
			}
		}

		if (NOW_S1_MODE) {
			// S1 mode
			crt_mon_s1vtaddr = ((*crtc_ma) >> (1 - width_sel)) + (width_sel & crt_mon_col);
			crt_mon_s1vgaddr = ((*crtc_ma) >> (1 - gwidth_sel)) + (gwidth_sel & crt_mon_col);
			if (SCRN_MODE_IS_640x400) {
				crt_mon_s1vgaddr += (((*crtc_ra) - (*crtc_odd_line)) & 0xf) << 11; // 0x800;
			} else {
				crt_mon_s1vgaddr += (((*crtc_ra) >> REG_INTERLACE_SEL) & 0x7) << (10 + gwidth_sel);	// 0x400 or 0x800 
			}
			// cursor
			cursor_addr = (((regs[14] << 8) | regs[15]) + (*disptmg_skew) - config.disptmg_skew);
			cursor_addr >>= (1 - width_sel); 
			cursor_addr &= 0x3fff;

		} else {
			// L3 mode
			addr_offset = hireso * (((*crtc_ra) >> REG_INTERLACE_SEL) & 0x7) << (10 + width_sel);	// 0x400 or 0x800
			crt_mon_l3vaddr = *crtc_ma;
			crt_mon_l3vaddr += crt_mon_col;
			crt_mon_l3vaddr += addr_offset;
			crt_mon_l3vaddr &= L3_VRAM_SIZE_1;
			if (crt_mon_l3vaddr < L3_ADDR_VRAM_START) {
				crt_mon_l3vaddr += L3_VRAM_SIZE;
			}

			// cursor
			cursor_addr = (((regs[14] << 8) | regs[15]) + (*disptmg_skew) - config.disptmg_skew);
			cursor_addr += addr_offset;
			cursor_addr &= 0x3fff;
			if (cursor_addr < L3_ADDR_VRAM_START) {
				cursor_addr += L3_VRAM_SIZE;
			}
		}

#ifdef DEBUG_CRTMON
		if (crt_mon_row < 2) {
			emu->out_debugf(_T("DISPLAY:update_vram: sw:%d crtc:cnt:%d ra:%d cmnr:%d cmr:%d cmsr:%d cmc:%d top:%d btm:%d vtop:%d vbtm:%d")
				,sw
				,(*crtc_vt_count), (*crtc_ra), crt_mon_now_raster
				,crt_mon_row, crt_mon_start_row,  crt_mon_col
				,top_raster, bottom_raster, vsync_top_raster, vsync_bottom_raster
			);
		}
#endif

		// pixel of width
		if (crt_mon_row < VRAM_BUF_HEIGHT) {
			dws_p_vline = &crt_mon_vline[crt_mon_row];

			dws_p_vline->raster = (*crtc_ra) << (1 - (*videomode));

			left_col = regs[0]-regs[2]-(regs[3] & 0x0f)-(crtc_chr_clocks ? 14 : 5) + (*disptmg_skew) - config.disptmg_skew;
			left_col *= crtc_ppch;
			left_col += ((SCREEN_WIDTH - 768) / 2);
			left_col += left_diff;

			dws_p_vline->left_col = left_col;
			dws_p_vline->right_col = left_col + crtc_ppch * regs[1];

			dws_p_vline->hsync_left_col = left_col + regs[2] * crtc_ppch;
			dws_p_vline->hsync_right_col = dws_p_vline->hsync_left_col + (regs[3] & 0x0f) * crtc_ppch;
		}
		// diff horizontal total display and crtc
		if (regs[0] < h_total_min[crtc_chr_clocks] || h_total_max[crtc_chr_clocks] < regs[0]) {
			left_diff += (regs[0] + 1 - (1 << (crtc_chr_clocks + 6))) * crtc_ppch;

			if (left_diff >= SCREEN_WIDTH) left_diff = -SCREEN_WIDTH + 1;
			else if (left_diff <= -SCREEN_WIDTH) left_diff = SCREEN_WIDTH - 1;
		}

		crt_mon_now_raster++;
		if (crt_mon_now_raster >= LINES_PER_FRAME) crt_mon_now_raster = 0;
	}

	now_skip_frame = emu->now_skip_frame();

	// copy vram data to buffer
	if (crt_mon_row < VRAM_BUF_HEIGHT && *crtc_vt_count < *crtc_vt_disp) {
		dws_p_vline = &crt_mon_vline[crt_mon_row];

		if (NOW_S1_MODE) {
			// S1 mode

			// save palette
			memcpy(dws_p_vline->pal, REG_PALETTE, sizeof(dws_p_vline->pal));

			// add page number 
			vtaddr = (crt_mon_s1vtaddr + (width_sel ? 0 : page_no) * (S1_VTRAM_SIZE >> 1));
			vtadiv = 1;
			vgaddr = (crt_mon_s1vgaddr + ((gwidth_sel || SCRN_MODE_IS_64COLOR) ? 0 : gpage_no) * (S1_VGRAM_SIZE_ONE >> 1));
			vgaddr2 = -1;
			vgadiv = 1;

			if (SCRN_MODE_IS_64COLOR) {
				// 64color mode
				vgaddr2 = vgaddr + S1_VGRAM_SIZE_HALF;
			} else if (SCRN_MODE_IS_640x400) {
				// 640x400 mode
				vgaddr2 = vgaddr + 0x800;
				vgaddr2 &= S1_VGRAM_SIZE_TWO_1;
				dws_p_vline->raster |= 0x80;
			}
			if (!now_skip_frame) {
			for(int i=0; i < VRAM_BUF_WIDTH; i += crt_mon_stepcols) {
				dws_p_crt_mon  = &crt_mon[crt_mon_row][crt_mon_col];

				if (vtaddr >= S1_VTRAM_SIZE) {
					vtaddr -= S1_VTRAM_SIZE;
				}
				if (SCRN_MODE_IS_640x400) {
					if (vgaddr >= S1_VGRAM_SIZE_TWO) {
						vgaddr -= S1_VGRAM_SIZE_TWO;
						vgaddr2 = vgaddr + 0x800;
					}
				} else {
					if (vgaddr >= S1_VGRAM_SIZE_ONE) {
						vgaddr -= S1_VGRAM_SIZE_ONE;
					}
				}
				if (crt_mon_col < regs[1]) {
					// text vram
					text = s1vtram[vtaddr];
					P_SET_CRT_MON(dws_p_crt_mon, text, text);
					// text color
					// bit 0-6: color ram  bit 7:width 0:40 1:80
					color = ((s1vcram[vtaddr] & 0x7f) | (width_sel << 7));

					P_SET_CRT_MON(dws_p_crt_mon, color, color);
					if (color & 0x20) {
						// ig
						addr = (text << 3)+((*crtc_ra) >> (*videomode));
						P_SET_CRT_MON(dws_p_crt_mon, c.b.b, ig_ram[addr]);
						P_SET_CRT_MON(dws_p_crt_mon, c.b.r, ig_ram[addr+IGRAM_SIZE]);
						P_SET_CRT_MON(dws_p_crt_mon, c.b.g, ig_ram[addr+IGRAM_SIZE+IGRAM_SIZE]);
						P_SET_CRT_MON(dws_p_crt_mon, c.b.i, 0);
					} else {
						P_SET_CRT_MON(dws_p_crt_mon, c.b.b, (color & 1) ? 0xff : 0);
						P_SET_CRT_MON(dws_p_crt_mon, c.b.r, (color & 2) ? 0xff : 0);
						P_SET_CRT_MON(dws_p_crt_mon, c.b.g, (color & 4) ? 0xff : 0);
						P_SET_CRT_MON(dws_p_crt_mon, c.b.i, 0);
					}
					// cursor
					if (vtaddr == cursor_addr && (*crtc_ra) <= 3) {
						if (crt_mon_col < crt_mon_cursor_col) crt_mon_cursor_col = crt_mon_col;
						if (crt_mon_row < crt_mon_cursor_row) crt_mon_cursor_row = crt_mon_row;
					}

					// graphic vram
					if (SCRN_MODE_IS_640x400) {
						// 640x400 mono
						if (*videomode) {
							// interlace only
							// even line
							P_SET_CRT_MON(dws_p_crt_mon, g[0].i, 0);
							if (vgaddr < S1_VGRAM_SIZE_ONE) {
								val = s1vgrram[vgaddr];
								P_SET_CRT_MON(dws_p_crt_mon, g[0].b.r, val);
							} else {
								val = s1vggram[vgaddr-S1_VGRAM_SIZE_ONE];
								P_SET_CRT_MON(dws_p_crt_mon, g[0].b.g, val);
							}
							// odd line
							P_SET_CRT_MON(dws_p_crt_mon, g[1].i, 0);
							if (vgaddr2 < S1_VGRAM_SIZE_ONE) {
								val2 = s1vgrram[vgaddr2];
								P_SET_CRT_MON(dws_p_crt_mon, g[1].b.r, val2);
							} else {
								val2 = s1vggram[vgaddr2-S1_VGRAM_SIZE_ONE];
								P_SET_CRT_MON(dws_p_crt_mon, g[1].b.g, val2);
							}
						} else {
							// non interlace ?
							P_SET_CRT_MON(dws_p_crt_mon, g[0].i, 0);
							P_SET_CRT_MON(dws_p_crt_mon, g[1].i, 0);
							if (vgaddr < S1_VGRAM_SIZE_ONE) {
								val = s1vgrram[vgaddr];
								P_SET_CRT_MON(dws_p_crt_mon, g[0].b.r, val);
								P_SET_CRT_MON(dws_p_crt_mon, g[1].b.r, val);
							} else {
								val = s1vggram[vgaddr-S1_VGRAM_SIZE_ONE];
								P_SET_CRT_MON(dws_p_crt_mon, g[0].b.g, val);
								P_SET_CRT_MON(dws_p_crt_mon, g[1].b.g, val);
							}
						}
						// attribute
						P_SET_CRT_MON(dws_p_crt_mon, brg_mask, disp_page_brg_mask);
					} else {
						// normal
						P_SET_CRT_MON(dws_p_crt_mon, g[0].b.r, s1vgrram[vgaddr]);
						P_SET_CRT_MON(dws_p_crt_mon, g[0].b.g, s1vggram[vgaddr]);
						P_SET_CRT_MON(dws_p_crt_mon, g[0].b.b, s1vgbram[vgaddr]);
						P_SET_CRT_MON(dws_p_crt_mon, g[0].b.i, 0);
						if (vgaddr2 >= 0) {
							// 320x200 64color
							P_SET_CRT_MON(dws_p_crt_mon, g[1].b.r, s1vgrram[vgaddr2]);
							P_SET_CRT_MON(dws_p_crt_mon, g[1].b.g, s1vggram[vgaddr2]);
							P_SET_CRT_MON(dws_p_crt_mon, g[1].b.b, s1vgbram[vgaddr2]);
							P_SET_CRT_MON(dws_p_crt_mon, g[1].b.i, 0);
						}
						// attribute
						P_SET_CRT_MON(dws_p_crt_mon, brg_mask, disp_page_brg_mask);
					}
					//   bit0-1:scrn_mode  bit5-7:disp graphic plane b g r
					P_SET_CRT_MON(dws_p_crt_mon, attr, ((DISP_PAGE_GRAPHIC_BRG) | (SCRN_MODE_GRAPHIC)));

				} else if (crt_mon_col < VRAM_BUF_WIDTH) {
					P_SET_CRT_MON(dws_p_crt_mon, text, 0);
					P_SET_CRT_MON(dws_p_crt_mon, color, 0);
					P_SET_CRT_MON(dws_p_crt_mon, g[0].i, 0);
					P_SET_CRT_MON(dws_p_crt_mon, g[1].i, 0);
					P_SET_CRT_MON(dws_p_crt_mon, attr, 0);
				}
				crt_mon_col += crt_mon_stepcols;
				vtadiv = crt_mon_stepcols - vtadiv;
				vtaddr += (width_sel ? crt_mon_stepcols : vtadiv);
				vgadiv = crt_mon_stepcols - vgadiv;
				vgaddr += (gwidth_sel ? crt_mon_stepcols : vgadiv);
				if (vgaddr2 >= 0) vgaddr2 += (gwidth_sel ? crt_mon_stepcols : vgadiv);
			}
			}
			crt_mon_s1vtaddr += vtaddr;
			crt_mon_s1vgaddr += vgaddr;
		} else {
			// L3 mode
			if (!now_skip_frame) {
			for(int i=0; i < (clock << width_sel); i += crt_mon_stepcols) {
				dws_p_crt_mon = &crt_mon[crt_mon_row][crt_mon_col];

				if (crt_mon_l3vaddr >= L3_ADDR_VRAM_END) {
					crt_mon_l3vaddr -= L3_VRAM_SIZE;
				}
				if (crt_mon_col < regs[1]) {
					P_SET_CRT_MON(dws_p_crt_mon, text, l3vram[crt_mon_l3vaddr]);
					// bit 0-5: color ram   bit 6: hireso
					P_SET_CRT_MON(dws_p_crt_mon, color, (color_ram[crt_mon_l3vaddr - L3_ADDR_VRAM_START] | (hireso << 6) | crt_mon_page) & (ig_mask ? 0xdf : 0xff));
					// cursor
					if (crt_mon_l3vaddr == cursor_addr && (*crtc_ra) <= 3) {
						if (crt_mon_col < crt_mon_cursor_col) crt_mon_cursor_col = crt_mon_col;
						if (crt_mon_row < crt_mon_cursor_row) crt_mon_cursor_row = crt_mon_row;
					}
				} else if (crt_mon_col < VRAM_BUF_WIDTH) {
					P_SET_CRT_MON(dws_p_crt_mon, text, 0);
					P_SET_CRT_MON(dws_p_crt_mon, color, 0);
				}
				crt_mon_col += crt_mon_stepcols;
				crt_mon_l3vaddr += crt_mon_stepcols;
			}
			}
		}
	}
}

// ----------------------------------------------------------------------------
#if 0
void DISPLAY::set_palette_brg()
{
	for(int i=0; i<8; i++) {
		set_palette_brg(i);
	}
}

void DISPLAY::set_palette_brg(int num)
{
	t_brg_color *p_brg = &palette_brg[num & 7];
	p_brg->b.b =  (REG_PALETTE[num] & 1);
	p_brg->b.r = ((REG_PALETTE[num] & 2) >> 1);
	p_brg->b.g = ((REG_PALETTE[num] & 4) >> 2);
	p_brg->b.i = ((REG_PALETTE[num] & 8) >> 3);
}
#endif

void DISPLAY::set_disp_page()
{
	page_no = (DISP_PAGE_TEXT) ? 1 : 0;
	gpage_no = (DISP_PAGE_GRAPHIC) ? 1 : 0;
	disp_page_brg_mask.b.b = (DISP_PAGE_GRAPHIC_B ? 0xff : 0);
	disp_page_brg_mask.b.r = (DISP_PAGE_GRAPHIC_R ? 0xff : 0);
	disp_page_brg_mask.b.g = (DISP_PAGE_GRAPHIC_G ? 0xff : 0);
	disp_page_brg_mask.b.i = 0xff;
}

void DISPLAY::set_s1_screen_mode()
{
	// for text
	width_sel = SCRN_MODE_TEXT_WIDTH ? 1 : 0;	// 0:w.40  1:w.80

	// for graphic
	gwidth_sel = SCRN_MODE_GRAPHIC_WIDTH ? 1 : 0;	// 640 or 320
}

void DISPLAY::change_palette()
{
	if (NOW_S1_MODE && FLG_ORIG_PAL8) {
		// dark gray
		COLOR16A(0,0,0,1) = emu->map_rgbcolor(0x1f,0x1f,0x1f);
	} else {
		// black 
		COLOR16A(0,0,0,1) = emu->map_rgbcolor(0x0,0x0,0x0);
	}
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

void DISPLAY::event_frame()
{
}

// ----------------------------------------------------------------------------

void DISPLAY::save_state(FILEIO *fio)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 0x41;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	vm_state.ig_enable = REG_IGENABLE;
	vm_state.ig_modereg = REG_IGMODE;
	vm_state.ig_mask = ig_mask ? 1 : 0;

	vm_state.sys_mode = REG_SYS_MODE;
	vm_state.mode_sel = REG_MODE_SEL;

	vm_state.chr_clocks = (crtc_chr_clocks & 0xff);
	vm_state.now_cpu_halt = now_cpu_halt;
	vm_state.interlace_sel = REG_INTERLACE_SEL;
	vm_state.tvsuper = REG_TVSUPER;

	//
	vm_state.bmsk_color = REG_BMSK_COLOR;
	vm_state.active_plane = REG_ACTIVE_PLANE;
	vm_state.dummy1 = 0;
	vm_state.disp_page = REG_DISP_PAGE;
	vm_state.scrn_mode = REG_SCRN_MODE;
	memcpy(vm_state.gcolor, REG_GCOLOR, sizeof(vm_state.gcolor));
	memcpy(vm_state.palette, REG_PALETTE, sizeof(vm_state.palette));

	fio->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fio->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void DISPLAY::load_state(FILEIO *fio)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;
	uint8 mode_sel = 0;

	READ_STATE_CHUNK(fio, vm_state_i, vm_state);

	REG_IGENABLE = vm_state.ig_enable;
	ig_mask = vm_state.ig_mask ? true : false;

	if (vm_state_i.version >= 2) {
		crtc_chr_clocks = vm_state.chr_clocks;
		crtc_ppch = (crtc_chr_clocks ? 8 : 16);
	}

	if (vm_state_i.version >= 0x41) {
		REG_SYS_MODE = vm_state.sys_mode;
		mode_sel = vm_state.mode_sel;

		REG_IGMODE = vm_state.ig_modereg;
		now_cpu_halt = vm_state.now_cpu_halt;
		REG_INTERLACE_SEL = vm_state.interlace_sel;
		REG_TVSUPER = vm_state.tvsuper;

		REG_BMSK_COLOR = vm_state.bmsk_color;
		REG_ACTIVE_PLANE = vm_state.active_plane;
		REG_DISP_PAGE = vm_state.disp_page;
		REG_SCRN_MODE = vm_state.scrn_mode;
		memcpy(REG_GCOLOR, vm_state.gcolor, sizeof(REG_GCOLOR));
		memcpy(REG_PALETTE, vm_state.palette, sizeof(REG_PALETTE));

		SIG_BMSK_DW = (BMSK_COLOR_BMSK ? (SIG_BMSK_DW | 2) : (SIG_BMSK_DW & ~2));
		SIG_BMSK_DW = (ACTIVE_PLANE_DW ? (SIG_BMSK_DW | 1) : (SIG_BMSK_DW & ~1));

		set_disp_page();

		if (NOW_S1_MODE) {
			set_s1_screen_mode();
		}
//		set_palette_brg();

		write_io8(0xffd0, mode_sel);
	}

	// clear
	crt_mon_row = 9999;
	crt_mon_col = 0;

}

