/** @file memory.cpp

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ memory ]
*/

#include <time.h>
#include "memory.h"
#include "../../fileio.h"
#include "../mc6809dasm.h"

#ifdef _DEBUG
//#define _DEBUG_RAM
//#define _DEBUG_CRAM
//#define _DEBUG_IO_COMMON
#define _DEBUG_TRACE_COUNTER
#endif

#define IGRAM_SIZE				0x800	// 2KB
#define IGRAM_SIZE_1			0x7ff	// 2KB-1
#define FONT_SIZE_1				0xfff	// 4KB-1

#define S1_ADDR_IGRAM_BANK1		0xbd000
#define S1_ADDR_IGRAM_BANK2		0xbe000
#define S1_ADDR_IGRAM_START		0xfa000
#define S1_ADDR_IGRAM_END		0xfa800

#define L3_ADDR_IGRAM_START		0xfa000
#define L3_ADDR_IGRAM_END		0xfa800

#define S1_ADDR_TRAP			0xbf000

#define VCRAM(addr) vcram[addr & S1_VTRAM_SIZE_1]
#define VGRAM_R(addr) ram[addr & S1_VGRAM_SIZE_ONE_1]
#define VGRAM_G(addr) vgram[addr & S1_VGRAM_SIZE_ONE_1]
#define VGRAM_B(addr) vgram[(addr & S1_VGRAM_SIZE_ONE_1)+S1_VGRAM_SIZE_ONE]
#define IGRAM(addr,brg) ig_ram[(addr & IGRAM_SIZE_1)+(IGRAM_SIZE*brg)]
#define FONTRAM(addr) font[addr & FONT_SIZE_1]

#define TRACE_COUNTER_PHASE_IDLE		0
#define TRACE_COUNTER_PHASE_WAITING		1
#define TRACE_COUNTER_PHASE_FIRED		2

void MEMORY::SET_BANK(int s, int e, uint8* w, uint8* r, uint8** wbank, uint8** rbank, int bank_size, uint8 *waitbank, uint8 waitval) {
	int sb = (s) >> bank_size, eb = (e) >> bank_size;
	for(int i = sb; i <= eb; i++) {
		if((w) == wdmy) {
			wbank[i] = wdmy;
		} else if ((w) != NULL) {
			wbank[i] = (w) + (1 << bank_size) * (i - sb);
		}
		if ((w) != NULL) {
			waitbank[i] = (waitval & 0xf0) | (waitbank[i] & 0x0f);
		}
		if((r) == rdmy) {
			rbank[i] = rdmy;
		} else if ((r) != NULL) {
			rbank[i] = (r) + (1 << bank_size) * (i - sb);
		}
		if ((r) != NULL) {
			waitbank[i] = (waitval & 0x0f) | (waitbank[i] & 0xf0);
		}
	}
}
#define L3_SET_BANK(s, e, w, r, wv) SET_BANK(s, e, w, r, l3wbank, l3rbank, L3_BANK_SIZE, l3wait, wv)
#define S1_SET_BANK(s, e, w, r, wv) SET_BANK(s, e, w, r, s1wbank, s1rbank, S1_BANK_SIZE, s1wait, wv)

void MEMORY::initialize()
{
	l3rom_loaded = false;
	s1rom1_loaded = false;
	s1rom2_loaded = false;
	s1dicrom_loaded = false;
	rom1802_loaded = false;
	rom1805_loaded = false;
	rom_loaded_at_first = false;

	memset(l3rom, 0xff, sizeof(l3rom));
	memset(s1rom, 0xff, sizeof(s1rom));
	memset(s1dicrom, 0xff, sizeof(s1dicrom));
	memset(rdmy, 0xff, sizeof(rdmy));
	memset(rom1802, 0xff, sizeof(rom1802));
	memset(rom1805, 0xff, sizeof(rom1805));

	memset(addr_map, 0, sizeof(addr_map));

	exram_size = 0;
	exram = NULL;

	// read rom image from file
	load_rom_files();

	// clear trace counter
//	trace_counter_id = -1;

	// allocate physical address
	for(int i=0; i<(0x100000 >> S1_BANK_SIZE); i++) {
		s1wbank[i] = wdmy;
		s1rbank[i] = rdmy;
	}
	memset(s1wait, 0, sizeof(s1wait));

	// init memory map
	reset_s1_membank();
	reset_l3_membank();

//	cpu_wait = 0;

	update_config();
}

void MEMORY::reset()
{
	// read rom image from file
	load_rom_files();

	// s1 or l3 mode
	reset_s1_l3_mode();

	memset(ram , 0, sizeof(ram));

	memset(vgram , 0, sizeof(vgram));
	reset_sram(vtram , sizeof(vtram));
	reset_sram(vcram , sizeof(vcram));

	// clear color ram
	srand((unsigned int)time(NULL));
	int co = (rand() % 3) * 3 + 1;	// 1 4 7
	int cg = (rand() % 2) * 24;		// 0 24
	color_reg = co + (co == 7 ? cg : 0);

	// clear address map
	memset(addr_map, 0, sizeof(addr_map));

	// reallocate extended ram
	set_extended_ram();

	// clear ig ram
	reset_igram();

	// init memory map
	reset_s1_membank();
	reset_l3_membank();

	// stop trace counter
	reset_trace_counter();

#ifdef USE_RTC
	rtc_reg_sel = 0;
	rtc_buf = 0;
#endif
}

void MEMORY::release()
{
	delete [] exram;
}

void MEMORY::load_rom_files()
{
	// load rom
	const _TCHAR *app_path, *rom_path[2];
	_TCHAR file_path[_MAX_PATH];

	rom_path[0] = config.rom_path;
	rom_path[1] = emu->application_path();

	FILEIO* fio = new FILEIO();
	for(int i=0; i<2; i++) {
		app_path = rom_path[i];

		if (!l3rom_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sL3BAS.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(l3rom, 0x6000, 1);
				fio->Fclose();
				l3rom_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("L3BAS.ROM was loaded."));
			}
		}
		if (!l3rom_loaded) {
			// try to load another main rom.
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sROM1.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(l3rom, 0x5f00, 1);
				fio->Fclose();
				emu->out_log(LOG_INFO, (CT)_("ROM1.ROM was loaded."));
			}
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sROM2.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(&l3rom[0x5ff0], 16, 1);
				fio->Fclose();
				l3rom_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("ROM2.ROM was loaded."));
			}
		}
		if (!s1rom1_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sS1BAS1.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(s1rom, 0x8000, 1);
				fio->Fclose();
				s1rom1_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("S1BAS1.ROM was loaded."));
			}
		}
		if (!s1rom2_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sS1BAS2.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(&s1rom[0x8000], 0x8000, 1);
				fio->Fclose();
				s1rom2_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("S1BAS2.ROM was loaded."));
			}
		}
		if (!s1rom2_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sS1ROM2.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(&s1rom[0x8000], 0x7e00, 1);
				fio->Fclose();
				emu->out_log(LOG_INFO, (CT)_("S1ROM2.ROM was loaded."));
			}
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sS1ROMI.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(&s1rom[0xfff0], 16, 1);
				fio->Fclose();
				s1rom2_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("S1ROMI.ROM was loaded."));
			}
		}
		if (!s1dicrom_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sS1DIC.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(s1dicrom, 0x8000, 1);
				fio->Fclose();
				s1dicrom_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("S1DIC.ROM was loaded."));
			}
		}
		if (!rom1802_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sMP1802.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(rom1802, 0x800, 1);
				fio->Fclose();
				rom1802_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("MP1802.ROM was loaded."));
			}
		}
		if (!rom1805_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sMP1805.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(rom1805, 0x800, 1);
				fio->Fclose();
				rom1805_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("MP1805.ROM was loaded."));
			}
		}
	}
	delete fio;

	if (!rom_loaded_at_first) {
		if (!l3rom_loaded) {
			emu->out_log(LOG_WARN, (CT)_("L3BAS.ROM couldn't be loaded."));
			emu->out_log(LOG_WARN, (CT)_("ROM1.ROM couldn't be loaded."));
			emu->out_log(LOG_WARN, (CT)_("ROM2.ROM couldn't be loaded."));
		}
		if (!s1rom1_loaded) {
			emu->out_log(LOG_WARN, (CT)_("S1BAS1.ROM couldn't be loaded."));
		}
		if (!s1rom2_loaded) {
			emu->out_log(LOG_WARN, (CT)_("S1BAS2.ROM couldn't be loaded."));
			emu->out_log(LOG_WARN, (CT)_("S1ROM2.ROM couldn't be loaded."));
			emu->out_log(LOG_WARN, (CT)_("S1ROMI.ROM couldn't be loaded."));
		}
		if (!s1dicrom_loaded) {
			emu->out_log(LOG_WARN, (CT)_("S1DIC.ROM couldn't be loaded."));
		}
		if (!rom1802_loaded) {
			emu->out_log(LOG_WARN, (CT)_("MP1802.ROM couldn't be loaded."));
		}
		if (!rom1805_loaded) {
			emu->out_log(LOG_WARN, (CT)_("MP1805.ROM couldn't be loaded."));
		}
		rom_loaded_at_first = true;
	}
}

void MEMORY::reset_sram(uint8 *sram, size_t size)
{
	for(size_t i=0; i<size; i+=2) {
		sram[i]=(rand() & 0xff);
		sram[i+1]=0xff;
	}
}

void MEMORY::reset_igram()
{
	// clear ig ram data
	for(int co=0; co<3; co++) {
		for(int ch=0; ch<256; ch++) {
			for(int i=0; i<8; i+=2) {
				ig_ram[(co*256+ch)*8+i]  =0x00;
				ig_ram[(co*256+ch)*8+i+1]=0xff;
				if ((ch % 3) == co) {
					switch(i) {
						case 0:
							if ((ch >> 4) < 8) ig_ram[(co*256+ch)*8+i]=(1 << (ch >> 4));
							break;
						case 2:
							if ((ch >> 4) < 16) ig_ram[(co*256+ch)*8+i]=(1 << ((ch >> 4) - 8));
							break;
						case 4:
							if ((ch & 15) < 8) ig_ram[(co*256+ch)*8+i]=(1 << (ch & 15));
							break;
						case 6:
							if ((ch & 15) < 16) ig_ram[(co*256+ch)*8+i]=(1 << ((ch & 15) - 8));
							break;
					}
				}
			}
		}
	}
//	ig_enable = 0;
	ig_enreg = 0;
}

void MEMORY::reset_s1_membank()
{
	// main ram 0x84000 - 0x8ffff
	S1_SET_BANK(0x84000, 0x8ffff, ram + 0x4000, ram + 0x4000, 0x77);
	// graphic ram 0xb0000 - 0xb3fff
	S1_SET_BANK(0xb0000, 0xb3fff, ram, ram, 0x77);
	// graphic ram 0xb4000 - 0xb7fff,0xb8000 - 0xbbfff (use in text only mode)
	S1_SET_BANK(0xb4000, 0xbbfff, vgram, vgram, 0x77);
	// text ram 0xbc000 - 0xbc7ff
	S1_SET_BANK(0xbc000, 0xbc7ff, vtram, vtram, 0x55);
//	// font rom 0xbd000 - 0xbdfff
//	S1_SET_BANK(0xbd000, 0xbdfff, wdmy, font);
	// dictionary rom 0xd0000 - 0xd7fff
	if (s1dicrom_loaded) S1_SET_BANK(0xd0000, 0xd7fff, wdmy, s1dicrom, 0);
	// basic rom 0xe0000 - 0xefdff
	// system io 0xefe00 - 0xeffff
	if (s1rom1_loaded) S1_SET_BANK(0xe0000, 0xe7fff, wdmy, s1rom,          0);
	if (s1rom2_loaded) S1_SET_BANK(0xe8000, 0xeffff, wdmy, s1rom + 0x8000, 0);
	// main ram (image) 0xf0000 - 0xfffff (unused)
	S1_SET_BANK(0xf0000, 0xfffff, ram, ram, 0x77);
	// l3 basic rom 0xfa000 - 0xfffff (unused)
	if (l3rom_loaded) S1_SET_BANK(0xfa000, 0xfffff, ram, l3rom, 0x70);
}

void MEMORY::reset_l3_membank()
{
	// init memory map
	mem_bank_reg1 = 0x0e;
	mem_bank_mask = true;
	mem_bank_reg2 = 0;
//	vram_sel = 0;
	change_l3_memory_bank();
}

void MEMORY::reset_trace_counter()
{
	// stop trace counter
//	if (trace_counter_id != -1) {
//		cancel_event(trace_counter_id);
//	}
//	trace_counter_id = -1;
	d_board->write_signal(SIG_CPU_NMI, 0, SIG_NMI_TRACE_MASK);
//	pc = 0;
	trace_counter_phase = TRACE_COUNTER_PHASE_IDLE;
	trace_count_remain = 0;
	prev_trace_clock = 0;

	// reset fuse
	REG_FUSE = 0x8f;
}

void MEMORY::set_igmode(uint32 data)
{
//	ig_enable = data & 0x01;
	d_disp->write_io8(0xffe9, data);
}

#define set_s1_cpu_wait(addr_1, addr_seg04_1, addr_bank_1, addr_comio_1, ref_mask, disp_mask, wait) \
{ \
	if (NOW_2MHZ) { \
		if (((*wait) % 30) == 0) { \
			/* dram refresh */ \
			if ( \
				(0xeff00 <= addr_comio_1 && addr_comio_1 <= 0xeffef) \
				|| (!config.mem_nowait && ( \
					(addr_1 >= 0xefe25 && addr_1 <= 0xefe27) \
				 || (s1wait[addr_bank_1] & ref_mask) \
				 || (addr_seg04_1 == S1_ADDR_TRAP) \
				   )) \
			) { \
				(*wait)++; \
				if (trace_counter_phase == TRACE_COUNTER_PHASE_WAITING) trace_count_remain++; \
			} \
		} else if (((*wait) & 1) == 0) { \
			/* type A wait */ \
			if ( \
				(0xeff00 <= addr_comio_1 && addr_comio_1 <= 0xeffef) \
				|| (!config.mem_nowait && ( \
					(addr_1 >= 0xefe25 && addr_1 <= 0xefe27) \
				 || (DISPTMG_ON && DISP_PAGE_GRAPHIC_BRG && (s1wait[addr_bank_1] & disp_mask)) \
				 || (addr_seg04_1 == S1_ADDR_TRAP) \
				   )) \
			) { \
				(*wait)++; \
				if (trace_counter_phase == TRACE_COUNTER_PHASE_WAITING) trace_count_remain++; \
			} \
		} else { \
			/* type B wait */ \
			if ( \
				(0xeff00 <= addr_comio_1 && addr_comio_1 <= 0xeffef) \
				|| (!config.mem_nowait && ( \
					(addr_1 >= 0xefe25 && addr_1 <= 0xefe27) \
				   )) \
			) { \
				(*wait)+=2; \
				if (trace_counter_phase == TRACE_COUNTER_PHASE_WAITING) trace_count_remain+=2; \
			} \
		} \
	} \
}

#define set_l3_cpu_wait(addr_1, addr_bank_1, addr_comio_1, ref_mask, disp_mask, wait) \
{ \
	if (NOW_2MHZ) { \
		if (((*wait) % 30) == 0) { \
			/* dram refresh */ \
			if ( \
				(0xeff00 <= addr_comio_1 && addr_comio_1 <= 0xeffef) \
				|| (!config.mem_nowait && ( \
					(l3wait[addr_bank_1] & ref_mask) \
				   )) \
			) { \
				(*wait)++; \
				if (trace_counter_phase == TRACE_COUNTER_PHASE_WAITING) trace_count_remain++; \
			} \
		} else if (((*wait) & 1) == 0) { \
			/* type A wait */ \
			if ( \
				(0xeff00 <= addr_comio_1 && addr_comio_1 <= 0xeffef) \
				|| (!config.mem_nowait && ( \
					(DISPTMG_ON && (l3wait[addr_bank_1] & disp_mask)) \
				   )) \
			) { \
				(*wait)++; \
				if (trace_counter_phase == TRACE_COUNTER_PHASE_WAITING) trace_count_remain++; \
			} \
		} else { \
			/* type B wait */ \
			if ( \
				(0xeff00 <= addr_comio_1 && addr_comio_1 <= 0xeffef) \
			) { \
				(*wait)+=2; \
				if (trace_counter_phase == TRACE_COUNTER_PHASE_WAITING) trace_count_remain+=2; \
			} \
		} \
	} \
}

void MEMORY::write_data8w(uint32 addr, uint32 data, int *wait)
{
	addr = address_mapping(addr & 0xffff);
	addr_seg04 = (addr & 0xff000); // 4KB  block
	addr_seg16 = (addr & 0xfc000); // 16KB block
	addr_seg32 = (addr & 0xf8000); // 32KB block
	addr_comio = (addr & 0xeffff);

#ifdef EMABLE_DISASSEMBLER
	// for debug
	dasm.set_phymem(addr, data & 0xff, true);
#endif

	if (addr < 0xf0000) {
		// S1 mode
		addr_bank = (addr >> S1_BANK_SIZE);

		// if access io port, cpu speed go down to 1MHz
		set_s1_cpu_wait(addr, addr_seg04, addr_bank, addr_comio, 0x30, 0x50, wait);

#ifdef _DEBUG_IO_COMMON
		if ((addr & 0xfff00) == 0xeff00) {
			uint8 ch = (data < 0x20 || data > 0x7f) ? 0x20 : data;
			emu->out_debugf("mw %04x=%02x %c",addr,data,ch);
		}
#endif
		// text vram
		if (S1_ADDR_VTRAM_START <= addr && addr < S1_ADDR_VTRAM_END) {
			// copy color register to color ram (7bit)
			VCRAM(addr) = (color_reg & 0x7f);
#ifdef _DEBUG_CVRAM
			if ((addr & 0x3ff) >= 0x0 && (addr & 0x3ff) < 0x28) {
				uint8 ch = (data < 0x20 || data > 0x7f) ? 0x20 : data;
				emu->out_debugf(_T("text w: %05x %02x %c c:%02x"),addr,data,ch,color_reg);
			}
#endif
		}
		// graphic vram
		if (DISP_PAGE_GRAPHIC_BRG != 0 && addr_seg16 == S1_ADDR_VGRAM_START) {
			if ((SIG_BMSK_DW & 3) == 0) {
				// direct write
				if (ACTIVE_PLANE_R) VGRAM_R(addr) = (data & 0xff);
				if (ACTIVE_PLANE_G) VGRAM_G(addr) = (data & 0xff);
				if (ACTIVE_PLANE_B) VGRAM_B(addr) = (data & 0xff);
			} else if ((SIG_BMSK_DW & 3) == 1) {
				// indirect write
				if (ACTIVE_PLANE_R) VGRAM_R(addr) = REG_GCOLOR_R;
				if (ACTIVE_PLANE_G) VGRAM_G(addr) = REG_GCOLOR_G;
				if (ACTIVE_PLANE_B) VGRAM_B(addr) = REG_GCOLOR_B;
			} else if ((SIG_BMSK_DW & 2) == 2) {
				// indirect write with bit mask
#if 1
				uint8 data_rev = ((~data) & 0xff);
				if (ACTIVE_PLANE_R) VGRAM_R(addr) = ((BMSK_COLOR_R ? (data & 0xff) : 0) | (REG_GCOLOR_R & data_rev));
				if (ACTIVE_PLANE_G) VGRAM_G(addr) = ((BMSK_COLOR_G ? (data & 0xff) : 0) | (REG_GCOLOR_G & data_rev));
				if (ACTIVE_PLANE_B) VGRAM_B(addr) = ((BMSK_COLOR_B ? (data & 0xff) : 0) | (REG_GCOLOR_B & data_rev));
#else
				if (ACTIVE_PLANE_R) VGRAM_R(addr) = 0;
				if (ACTIVE_PLANE_G) VGRAM_G(addr) = 0;
				if (ACTIVE_PLANE_B) VGRAM_B(addr) = 0;
				for(uint8 dot=0x80; dot>0; dot >>= 1) {
					if (ACTIVE_PLANE_R) {
						if (data & dot) VGRAM_R(addr) |= (BMSK_COLOR_R ? dot : 0);
						else 			VGRAM_R(addr) |= (REG_GCOLOR_R & dot);
					}
					if (ACTIVE_PLANE_G) {
						if (data & dot) VGRAM_G(addr) |= (BMSK_COLOR_G ? dot : 0);
						else			VGRAM_G(addr) |= (REG_GCOLOR_G & dot);
					}
					if (ACTIVE_PLANE_B) {
						if (data & dot) VGRAM_B(addr) |= (BMSK_COLOR_B ? dot : 0);
						else			VGRAM_B(addr) |= (REG_GCOLOR_B & dot);
					}
				}
#endif
			}
#if 0	// color ram is not modified.
#ifdef _DEBUG_CVRAM
			if ((addr & 0x3ff) >= 0x0 && (addr & 0x3ff) < 0x28) {
				uint8 ch = (data < 0x20 || data > 0x7f) ? 0x20 : data;
				emu->out_debugf(_T("graphic w: %05x %02x %c c:%02x"),addr,data,ch,color_reg);
			}
#endif
			// copy color register to color ram (7bit)
			if (SCRN_MODE_T_AND_G_WIDTH == 0x06) {
				// gw.640 and w.80
				vcram[addr & 0x7ff] = (color_reg & 0x7f);
			} else if (SCRN_MODE_T_AND_G_WIDTH == 0x02) {
				// gw.640 and w.40
				vcram[(addr & 0x7ff) >> 1] = (color_reg & 0x7f);
			} else if (SCRN_MODE_T_AND_G_WIDTH == 0x04) {
				// gw.320 and w.80
				vcram[(addr & 0x3ff) >> 1] = (color_reg & 0x7f);
			} else {
				// gw.320 and w.40
				vcram[addr & 0x3ff] = (color_reg & 0x7f);
			}
#endif
			return;
		}
		// ig ram
		if (addr_seg04 == S1_ADDR_IGRAM_BANK1 || addr_seg04 == S1_ADDR_IGRAM_BANK2) {
			for(int i=0; i<3; i++) {
				if (ig_enreg & (1 << i)) {
					IGRAM(addr,i) = (data & 0xff);
				}
			}
			return;
		}
		s1wbank[addr_bank][addr & ((1 << S1_BANK_SIZE) - 1)] = data;
	} else {
		// L3 mode or upper 0xf0000
		addr_64kb = (addr & 0xffff);
		addr_bank = (addr_64kb >> L3_BANK_SIZE);

		// if access io port, cpu speed go down to 1MHz
		set_l3_cpu_wait(addr, addr_bank, addr_comio, 0x30, 0x50, wait);

		// vram area
		if (L3_ADDR_VRAM_START <= addr_64kb && addr_64kb < L3_ADDR_VRAM_END) {
#ifdef _DEBUG_CRAM
//			if ((addr - ADDR_VRAM_START) < 0x0400) {
				uint8 ch=data;
				ch = (ch < 0x20 || ch > 0x7f) ? 0x20 : ch;
				emu->out_debugf("mw %04x=%02x->%02x %c c%02x->%02x",addr,ram[addr],data,ch,color_reg,vgram[addr]);
//			}
#endif
			vgram[(addr_64kb - L3_ADDR_VRAM_START) & L3_VRAM_SIZE_1] = (color_reg & 0x3f);
		}
		// ig_area
		if (REG_IGENABLE) {
			if (L3_ADDR_IGRAM_START <= addr && addr < L3_ADDR_IGRAM_END) {
				for(int i=0; i<3; i++) {
					if (ig_enreg & (1 << i)) {
						IGRAM(addr,i) = data & 0xff;
					}
				}
				return;
			}
		}
		l3wbank[addr_bank][addr & ((1 << L3_BANK_SIZE) - 1)] = data;
	}

	// memory mapped i/o (s1 only area : 0xefe00 - 0xefeff)
	switch(addr & 0xfffff) {
		case 0xefe00:
		case 0xefe01:
		case 0xefe02:
		case 0xefe03:
		case 0xefe04:
		case 0xefe05:
		case 0xefe06:
		case 0xefe07:
		case 0xefe08:
		case 0xefe09:
		case 0xefe0a:
		case 0xefe0b:
		case 0xefe0c:
		case 0xefe0d:
		case 0xefe0e:
		case 0xefe0f:
			// address mapping
			if (NOW_SYSTEM_MODE) {
				addr_map[REG_ADDRSEG][addr & 0xf] = data & 0xff;
			}
			break;
		case 0xefe10:
			// fuse system / user mode
			if (NOW_SYSTEM_MODE) {
				REG_FUSE = data & 0xcf;
				d_board->write_signal(SIG_FUSE_INTR_MASK, data, FUSE_INTR_MASK);
#ifdef _DEBUG_TRACE_COUNTER
				emu->out_debugf(_T("fuse w: 0x%02x"), REG_FUSE);
#endif
			}
			break;
		case 0xefe11:
			// address segment
			if (IOPORT_USE_OS9BD && NOW_SYSTEM_MODE) {
				REG_ADDRSEG = data & 0x3f;
			}
			break;
		case 0xefe18:
			// trap flag (cannot write)
			break;
		case 0xefe19:
			// bus control
			REG_BUSCTRL = (data & BUSCTRL_SIGNAL);
			break;
		case 0xefe20:
		case 0xefe21:
		case 0xefe22:
		case 0xefe23:
		case 0xefe24:
		case 0xefe25:
		case 0xefe26:
		case 0xefe27:
		case 0xefe28:
		case 0xefe29:
		case 0xefe2a:
		case 0xefe2b:
		case 0xefe2c:
		case 0xefe2d:
		case 0xefe2e:
		case 0xefe2f:
			// graphic related register
			d_disp->write_io8(addr, data);
			break;
		case 0xefe40:
		case 0xefe41:
			// pia a port (joystick)
			d_pia->write_io8(addr - 0xefe40, data);
			break;
		case 0xefe42:
		case 0xefe43:
			// pia b (printer)
			d_pia->write_io8(addr - 0xefe40, data);
			break;
	}

	// memory mapped i/o (common area : 0xeff00 - 0xeffef, 0xffe00 - 0xfffef)
	switch(addr & 0xeffff) {
#ifdef USE_FD1
		case 0xeff00:
		case 0xeff01:
		case 0xeff02:
		case 0xeff03:
			// fdc
		case 0xeff04:
			// fdd drive select
			if (IOPORT_USE_5FDD) {
				d_fdd->write_io8(addr, data);
			}
			break;
		case 0xeff08:
			// halt (1MB fdd only)
			if (IOPORT_USE_5FDD) {
				d_fdd->write_io8(addr, data);
			}
			break;
		case 0xeff0c:
			// type sel (cannot write)
			break;
		case 0xeff10:
		case 0xeff11:
		case 0xeff12:
		case 0xeff13:
			// fdc (ex)
		case 0xeff14:
			// fdd (ex) drive select
			if (IOPORT_USE_5FDD) {
				d_fdd->write_io8(addr, data);
			}
			break;
		case 0xeff18:
		case 0xeff19:
		case 0xeff1a:
		case 0xeff1b:
		case 0xeff1c:
		case 0xeff1d:
		case 0xeff1e:
		case 0xeff1f:
			// fdc
			if (IOPORT_USE_3FDD) {
				d_fdc3->write_io8(addr - 0xeff18, data);

//				emu->out_debugf("fdc3w a:%04x d:%02x",addr,data);
			}
			break;
		case 0xeff20:
			// fdd drive select
			if (IOPORT_USE_3FDD) {
				d_fdd->write_io8(0, data);

//				emu->out_debugf("fdd3w a:%04x d:%02x",addr,data);
			}
			break;
#endif
		case 0xeff30:
		case 0xeff31:
		case 0xeff32:
		case 0xeff33:
			// psg exboard
			if (IOPORT_USE_PSG6) {
				d_psg3->write_io8(addr - 0xeff30, data);
//				emu->out_debugf("psgw a:%04x d:%02x",addr,data);
			}
			break;
#ifdef USE_RTC
		case 0xeff38:
			// rtc register select
			rtc_reg_sel = data & 0xff;
			break;

		case 0xeff39:
			// rtc write
			if ((data & 0x80) == 0) {
				rtc_buf = data & 0xff;
			}
			break;
#endif
		case 0xeff3c:
		case 0xeff3d:
		case 0xeff3e:
		case 0xeff3f:
			// extended LPT board
			if (IOPORT_USE_EXPIA) {
				d_pia_ex->write_io8(addr - 0xeff3c, data);
			}
			break;

		case 0xeff40:
			// extended COM board
			if (IOPORT_USE_EXACIA) {
				d_comm1->write_io8(addr - 0xeff40, data);
			}
			// through
		case 0xeff41:
			// extended COM board
			if (IOPORT_USE_EXACIA) {
				d_acia_ex->write_io8(addr - 0xeff40, data);
			}
			break;
		case 0xeff70:
		case 0xeff71:
		case 0xeff72:
		case 0xeff73:
		case 0xeff74:
		case 0xeff77:
		case 0xeff78:
		case 0xeff79:
		case 0xeff7a:
		case 0xeff7b:
		case 0xeff7c:
		case 0xeff7d:
		case 0xeff7e:
		case 0xeff7f:
			// psg9
			if (IOPORT_USE_PSG9) {
				d_psg9->write_io8(addr - 0xeff70, data);
//				emu->out_debugf("psg9w a:%04x d:%02x",addr,data);
			}
			break;
		case 0xeff75:
		case 0xeff76:
			// psg9
			if (IOPORT_USE_PSG9) {
				d_psg9->write_io8(addr - 0xeff70, data);
//				emu->out_debugf("psg9w a:%04x d:%02x",addr,data);
			}
			// kanji rom
			if (IOPORT_USE_KANJI) {
				d_kanji->write_io8(addr - 0xeff75, data);
//				emu->out_debugf("kanjiw a:%04x d:%02x",addr,data);
			}
			break;

		case 0xeffc0:
			// membank (L3 only)
			if (!mem_bank_mask) {
				mem_bank_reg1 = data & 0xce;
				change_l3_memory_bank();
			}
			break;

		case 0xeffc1:
			mem_bank_mask = ((data & 4) == 0);
			break;

		case 0xeffc2:
		case 0xeffc3:
			// pia b (printer)
			d_pia->write_io8(addr - 0xeffc0, data);
			break;

		case 0xeffc4:
			// acia
			d_comm->write_io8(addr - 0xeffc4, data);
			// pass through
		case 0xeffc5:
			// acia
			d_acia->write_io8(addr - 0xeffc4, data);
			break;

		case 0xeffc6:
		case 0xeffc7:
			// crtc
			d_crtc->write_io8(addr - 0xeffc6, data);
//			d_disp->write_io8(addr, data);
			break;

		case 0xeffc8:
			// keyborad nmi (cannot write)
			break;
		case 0xeffca:
			// timer irq
			break;
		case 0xeffcb:
			// light pen flag (cannot write)
			break;
		case 0xeffd0:
			// mode select
			REG_MODE_SEL = (data & 0xff);
			d_disp->write_io8(addr, data);
			d_comm->write_signal(SIG_COMM_RS, data, 0x20);
			if (NOW_L3_MODE) {
				d_crtc->write_signal(SIG_HD46505_CHR_CLOCKS, CLOCKS_1MHZ * ((data & 0x80) ? 2 : 1), 0xffffffff);
				d_disp->write_signal(SIG_DISPLAY_CHR_CLOCKS, CLOCKS_1MHZ * ((data & 0x80) ? 2 : 1), 0xffffffff);
			}
			break;
		case 0xeffd1:
			// trace counter
			fetch_trace_counter(*wait);
			break;
		case 0xeffd2:
			// remote switch
			d_cmt->write_signal(SIG_CMT_REMOTE, data, 0x80);
			break;
		case 0xeffd3:
			// music sel
//			emu->out_debugf("ms %02x",data);
			d_sound->write_signal(SIG_SOUND_ON, 1, 1);
			d_sound->write_signal(SIG_SOUND_SIGNAL, data, 0x80);
			break;
		case 0xeffd4:
			// time mask
			d_timer->write_io8(addr, data & 0x83);
			break;
		case 0xeffd5:
			// light pen bl
//			emu->out_debugf(_T("lpengl: %02x"),data);
			d_key->write_io8(addr, data & 0x80);
			break;
		case 0xeffd6:
			// interace sel
			d_disp->write_io8(addr, data & 0x08);
			break;
		case 0xeffd7:
			// baud sel
			d_cmt->write_signal(SIG_CMT_BAUD_SEL, data, 0x0f);
			break;
		case 0xeffd8:
			// color register
#ifdef _DEBUG_CRAM
			emu->out_debugf("cw c%02x->%02x",color_reg,data);
#endif
			color_reg = (data & 0xff);
			break;
		case 0xeffd9:
			// trq-sel casette control
			// not implemented
			break;
		case 0xeffdc:
		case 0xeffdd:
		case 0xeffde:
		case 0xeffdf:
			// mouse
			d_mouse->write_io8(addr, data);
			break;
		case 0xeffe0:
			// keyboard mode register
			d_key -> write_io8(addr, data);
			break;
		case 0xeffe1:
			// kb type (cannot write)
			break;
		case 0xeffe2:
			// tv super inpose
			REG_TVSUPER = (data & 0x17);
			break;
		case 0xeffe4:
			// psg data
			d_psg[0]->write_io8(1, data & 0xff);
//			emu->out_debugf("psgw dat:%02x",data);
			break;
		case 0xeffe5:
			// psg reg
			d_psg[0]->write_io8(0, data & 0x0f);
//			emu->out_debugf("psgw reg:%02x",data);
			break;
		case 0xeffe6:
			// ex psg data
			if (IOPORT_USE_EXPSG) {
				d_psg[1]->write_io8(1, data & 0xff);
//				emu->out_debugf("psg2w dat:%02x",data);
			}
			break;
		case 0xeffe7:
			// ex psg reg
			if (IOPORT_USE_EXPSG) {
				d_psg[1]->write_io8(0, data & 0x0f);
//				emu->out_debugf("psg2w reg:%02x",data);
			}
			break;
		case 0xeffe8:
			// bank register
			mem_bank_reg2 = data & 0xff;
			change_l3_memory_bank();
			break;
		case 0xeffe9:
			// IG mode register
			set_igmode(0x100 | data);
			break;
		case 0xeffea:
			// IG en register
			ig_enreg = data & 0x07;
			break;
		case 0xeffeb:
			// system mode
			set_cpu_clock(data & 0xff);
			break;
	}
	// send trap signal when access the address 0xbfxxx 
	if (addr_seg04 == S1_ADDR_TRAP) {
		REG_TRAPF |= TRAPF_SIGNAL;
		d_board->write_signal(SIG_CPU_NMI, SIG_NMI_TRAP_MASK, SIG_NMI_TRAP_MASK);
	}

}

uint32 MEMORY::read_data8w(uint32 addr, int *wait)
{
	uint32 data;

	addr = address_mapping(addr & 0xffff);
	addr_seg04 = (addr & 0xff000); // 4KB  block
	addr_seg16 = (addr & 0xfc000); // 16KB block
	addr_seg32 = (addr & 0xf8000); // 32KB block
	addr_comio = (addr & 0xeffff);

#ifdef EMABLE_DISASSEMBLER
	// for debug
	dasm.set_phyaddr(addr);
	dasm.set_phymem(addr, 0, false);
#endif

	if (addr < 0xf0000) {
		// S1 mode
		addr_bank = (addr >> S1_BANK_SIZE);

		// if access io port, cpu speed go down to 1MHz
		set_s1_cpu_wait(addr, addr_seg04, addr_bank, addr_comio, 0x03, 0x05, wait);

		// trace counter
		process_trace_counter(*wait);

		data = s1rbank[addr_bank][addr & ((1 << S1_BANK_SIZE) - 1)];

#ifdef EMABLE_DISASSEMBLER
		if (data == 0x3f && s1rbank[((addr-1) >> S1_BANK_SIZE) & 0x1ff][(addr-1) & ((1 << S1_BANK_SIZE) - 1)] == 0x10) {
			dasm.set_swi2_param(
				s1rbank[(addr + 1) >> S1_BANK_SIZE][(addr + 1) & ((1 << S1_BANK_SIZE) - 1)],
				s1rbank[(addr + 2) >> S1_BANK_SIZE][(addr + 2) & ((1 << S1_BANK_SIZE) - 1)]
			);
		}
#endif
		// text vram
		if (S1_ADDR_VTRAM_START <= addr && addr < S1_ADDR_VTRAM_END) {
			// 7bit
			// if MK bit is set, do not read from color ram.
			if ((color_reg & 0x80) == 0) {
				color_reg = (VCRAM(addr) & 0x7f);
			}
#ifdef _DEBUG_CVRAM
			if ((addr & 0x3ff) >= 0x0 && (addr & 0x3ff) < 0x28) {
				uint8 ch = (data < 0x20 || data > 0x7f) ? 0x20 : data;
				emu->out_debugf(_T("text r: %05x %02x %c c:%02x"),addr,data,ch,color_reg);
			}
#endif
			return data;
		}
		// graphic vram
		if (DISP_PAGE_GRAPHIC_BRG != 0 && addr_seg16 == S1_ADDR_VGRAM_START) {
			if ((SIG_BMSK_DW & 1) == 0) {
				// direct read
				data = 0;
				if (ACTIVE_PLANE_G) {
					data = VGRAM_G(addr);
				} else if (ACTIVE_PLANE_R) {
					data = VGRAM_R(addr);
				} else if (ACTIVE_PLANE_B) {
					data = VGRAM_B(addr);
				}
			} else if ((SIG_BMSK_DW & 1) == 1) {
				// indirect read
				data = 0;
				if (ACTIVE_PLANE_G) REG_GCOLOR_G = VGRAM_G(addr);
				if (ACTIVE_PLANE_R) REG_GCOLOR_R = VGRAM_R(addr);
				if (ACTIVE_PLANE_B) REG_GCOLOR_B = VGRAM_B(addr);
				if (ACTIVE_PLANE_G) {
					data = REG_GCOLOR_G;
				} else if (ACTIVE_PLANE_R) {
					data = REG_GCOLOR_R;
				} else if (ACTIVE_PLANE_B) {
					data = REG_GCOLOR_B;
				}
			}
#if 0
			// 7bit
			// if MK bit is set, do not read from color ram.
			if ((color_reg & 0x80) == 0) {
				if (SCRN_MODE_T_AND_G_WIDTH == 0x06) {
					// gw.640 and w.80
					color_reg = (vcram[addr & 0x7ff] & 0x7f);
				} else if (SCRN_MODE_T_AND_G_WIDTH == 0x02) {
					// gw.640 and w.40
					color_reg = (vcram[(addr & 0x7ff) >> 1] & 0x7f);
				} else if (SCRN_MODE_T_AND_G_WIDTH == 0x04) {
					// gw.320 and w.80
					color_reg = (vcram[(addr & 0x3ff) >> 1] & 0x7f);
				} else {
					// gw.320 and w.40
					color_reg = (vcram[addr & 0x3ff] & 0x7f);
				}
			}
#ifdef _DEBUG_CVRAM
			if ((addr & 0x3ff) >= 0x0 && (addr & 0x3ff) < 0x28) {
				uint8 ch = (data < 0x20 || data > 0x7f) ? 0x20 : data;
				emu->out_debugf(_T("graphic r: %05x %02x %c c:%02x"),addr,data,ch,color_reg);
			}
#endif
#endif
			return data;
		}
		// ig ram (or font rom)
		if (addr_seg04 == S1_ADDR_IGRAM_BANK1 || addr_seg04 == S1_ADDR_IGRAM_BANK2) {
			if ((ig_enreg & 0x07) == 0x07) {
				// readable font rom
				data = FONTRAM(addr);
			} else {
				data = 0;
				for(int i=2; i>=0; i--) {
					// G > R > B
					if (ig_enreg & (1 << i)) {
						data = IGRAM(addr,i);
						break;
					}
				}
			}
			return data;
		}
	} else {
		// L3 mode or upper 0xf0000
		addr_64kb = (addr & 0xffff);
		addr_bank = (addr_64kb >> L3_BANK_SIZE);

		// if access io port, cpu speed go down to 1MHz
		set_l3_cpu_wait(addr, addr_bank, addr_comio, 0x03, 0x05, wait);

		// trace counter
		process_trace_counter(*wait);

		// vram area
		if (L3_ADDR_VRAM_START <= addr_64kb && addr_64kb < L3_ADDR_VRAM_END) {
#ifdef _DEBUG_CRAM
//			if ((addr - ADDR_VRAM_START) < 0x0400) {
				uint8 ch=ram[addr];
				ch = (ch < 0x20 || ch > 0x7f) ? 0x20 : ch;
				emu->out_debugf("mr %04x=%02x %c c%02x r%02x",addr,ram[addr],ch,vgram[addr],color_reg);
//			}
#endif
			// 6bit
			// if MK bit is set, do not read from color ram.
			if ((color_reg & 0x80) == 0) {
				color_reg = (vgram[(addr_64kb - L3_ADDR_VRAM_START) & L3_VRAM_SIZE_1] & 0x3f);
			}
		}
		data = l3rbank[addr_bank][addr & ((1 << L3_BANK_SIZE) - 1)];
	}

	// memory mapped i/o (s1 only area : 0xefe00 - 0xefeff)
	switch(addr & 0xfffff) {
		case 0xefe00:
		case 0xefe01:
		case 0xefe02:
		case 0xefe03:
		case 0xefe04:
		case 0xefe05:
		case 0xefe06:
		case 0xefe07:
		case 0xefe08:
		case 0xefe09:
		case 0xefe0a:
		case 0xefe0b:
		case 0xefe0c:
		case 0xefe0d:
		case 0xefe0e:
		case 0xefe0f:
			// address mapping
			if (NOW_SYSTEM_MODE) {
				data = addr_map[REG_ADDRSEG][addr & 0xf];
			}
			break;
		case 0xefe10:
			// fuse (cannot read)
			if (NOW_SYSTEM_MODE) {
				data = 0xff;
			}
			break;
		case 0xefe11:
			// segment (cannot read)
			if (IOPORT_USE_OS9BD && NOW_SYSTEM_MODE) {
				data = 0xff;
			}
			break;
		case 0xefe18:
			// trap flag
			data = (REG_TRAPF & TRAPF_SIGNAL);
			REG_TRAPF &= ~TRAPF_SIGNAL;
			d_board->write_signal(SIG_CPU_NMI, 0, SIG_NMI_TRAP_MASK);
			break;
		case 0xefe19:
			// bus control
			data = (REG_BUSCTRL & BUSCTRL_SIGNAL);
			break;
		case 0xefe20:
			// bmsk color graphic color (cannot read)
			break;
		case 0xefe21:
			// active graphic plane (cannot read)
			break;
		case 0xefe23:
			// display page (cannot read)
			break;
		case 0xefe24:
			// screen page (cannot read)
			break;
		case 0xefe25:
		case 0xefe26:
		case 0xefe27:
			// graphic color
			data = REG_GCOLOR[addr - 0xefe25];
			break;
		case 0xefe28:
		case 0xefe29:
		case 0xefe2a:
		case 0xefe2b:
		case 0xefe2c:
		case 0xefe2d:
		case 0xefe2e:
		case 0xefe2f:
			// palette (cannot read, but send halt signal when disptmg is on)
			d_disp->read_io8(addr);
			break;
		case 0xefe40:
		case 0xefe41:
			// pia a port (joystick)
			data = d_pia->read_io8(addr - 0xefe40);
			break;
		case 0xefe42:
		case 0xefe43:
			// pia b (printer)
			data = d_pia->read_io8(addr - 0xefe40);
			break;
	}

	// memory mapped i/o (common area : 0xeff00 - 0xeffef, 0xfff00 - 0xfffef)
	switch(addr & 0xeffff) {
#ifdef USE_FD1
		case 0xeff00:
		case 0xeff01:
		case 0xeff02:
		case 0xeff03:
			// fdc
		case 0xeff04:
			// fdd drive select
			if (IOPORT_USE_5FDD) {
				data = d_fdd->read_io8(addr);
			}
			break;
		case 0xeff08:
			// now halt (1MB fdd only) (cannot read)
			break;
		case 0xeff0c:
			// type sel (1MB fdd only)
			if (IOPORT_USE_5FDD) {
				data = d_fdd->read_io8(addr);
			}
			break;
		case 0xeff10:
		case 0xeff11:
		case 0xeff12:
		case 0xeff13:
			// fdc (ex)
		case 0xeff14:
			// fdd (ex) drive select
			if (IOPORT_USE_5FDD) {
				data = d_fdd->read_io8(addr);
			}
			break;
		case 0xeff18:
		case 0xeff19:
		case 0xeff1a:
		case 0xeff1b:
		case 0xeff1c:
		case 0xeff1d:
		case 0xeff1e:
		case 0xeff1f:
			// fdc
			if (IOPORT_USE_3FDD) {
				data = d_fdc3->read_io8(addr - 0xeff18);

//				emu->out_debugf("fdc3r a:%04x d:%02x",addr,data);
			}
			break;
		case 0xeff20:
			// fdd drive select
			if (IOPORT_USE_3FDD) {
				data = d_fdd->read_io8(0);

//				emu->out_debugf("fdd3r a:%04x d:%02x",addr,data);
			}
			break;
#endif
		case 0xeff30:
		case 0xeff31:
		case 0xeff32:
		case 0xeff33:
			// psg exboard
			if (IOPORT_USE_PSG6) {
				data = d_psg3->read_io8(addr - 0xeff30);
//				emu->out_debugf("psgr a:%04x d:%02x",addr,data);
			}
			break;
#ifdef USE_RTC
		case 0xeff3a:
			// rtc read
			if ((rtc_reg_sel & 0xc0) == 0x80) {
				int tim[8];
				emu->get_timer(tim);
				switch (rtc_reg_sel & 0x0f) {
				case 0:
					data = (tim[6] % 10);
					break;
				case 1:
					data = (tim[6] / 10);
					break;
				case 2:
					data = (tim[5] % 10);
					break;
				case 3:
					data = (tim[5] / 10);
					break;
				case 4:
					data = (tim[4] % 10);
					break;
				case 5:
					data = (tim[4] / 10);
					break;
				case 6:
					data = (tim[3] % 10);
					break;
				case 7:
					data = (tim[2] % 10);
					break;
				case 8:
					data = (tim[2] / 10);
					break;
				case 9:
					data = (tim[1] % 10);
					break;
				case 10:
					data = (tim[1] / 10);
					break;
				case 11:
					data = (tim[0] % 10);
					break;
				case 12:
					data = ((tim[0] / 10) % 10);
					break;
				}
			} else {
				data = rtc_buf & 0xff;
			}
			break;
#endif
		case 0xeff3c:
		case 0xeff3d:
		case 0xeff3e:
		case 0xeff3f:
			// extended LPT board
			if (IOPORT_USE_EXPIA) {
				data = d_pia_ex->read_io8(addr - 0xeff3c);
			}
			break;

		case 0xeff40:
		case 0xeff41:
			// extended COM board
			if (IOPORT_USE_EXACIA) {
				data = d_acia_ex->read_io8(addr - 0xeff40);
			}
			break;
		case 0xeff70:
		case 0xeff71:
		case 0xeff72:
		case 0xeff73:
		case 0xeff74:
		case 0xeff77:
		case 0xeff78:
		case 0xeff79:
		case 0xeff7a:
		case 0xeff7b:
		case 0xeff7c:
		case 0xeff7d:
		case 0xeff7e:
		case 0xeff7f:
			// psg9
			if (IOPORT_USE_PSG9) {
				data = d_psg9->read_io8(addr - 0xeff70);
//				emu->out_debugf("psg9r a:%04x d:%02x",addr,data);
			}
			break;
		case 0xeff75:
		case 0xeff76:
			// psg9
			if (IOPORT_USE_PSG9) {
				data = d_psg9->read_io8(addr - 0xeff70);
//				emu->out_debugf("psg9r a:%04x d:%02x",addr,data);
			}
			// kanji rom
			if (IOPORT_USE_KANJI) {
				data = d_kanji->read_io8(addr - 0xeff75);
//				emu->out_debugf("kanjir a:%04x d:%02x",addr,data);
			}
			break;

		case 0xeffc0:
			// membank (L3 mode)
			data = mem_bank_reg1;
			break;

		case 0xeffc1:
			break;

		case 0xeffc2:
		case 0xeffc3:
			// pia b (printer)
			data = d_pia->read_io8(addr - 0xeffc0);
			break;

		case 0xeffc4:
		case 0xeffc5:
			// acia
			data = d_acia->read_io8(addr - 0xeffc4);
			break;

		case 0xeffc6:
		case 0xeffc7:
			// crtc
			data = d_crtc->read_io8(addr - 0xeffc6);
			break;

		case 0xeffc8:
			// keyboard nmi
			data = d_key->read_io8(addr);
			break;
		case 0xeffc9:
			// newon number
			data = config.dipswitch;
			break;
		case 0xeffca:
			// timer irq
			data = d_timer->read_io8(addr);
			break;
		case 0xeffcb:
			// light pen flag 
			data = d_key->read_io8(addr);
			break;
		case 0xeffd0:
			// mode select
			break;
		case 0xeffd1:
			// trace counter
			break;
		case 0xeffd2:
			// remote switch
			break;
		case 0xeffd3:
			// music sel
			break;
		case 0xeffd4:
			// time mask
			break;
		case 0xeffd5:
			// light pen bl
			break;
		case 0xeffd6:
			// interace sel
			break;
		case 0xeffd7:
			// baud sel
			break;
		case 0xeffd8:
			// color register sel
#ifdef _DEBUG_CRAM
			emu->out_debugf("cr c%02x",color_reg);
#endif
			data = (color_reg & 0xff);
			break;
		case 0xeffd9:
			// trq-sel casette control
			data = 0xf;
			break;
		case 0xeffdc:
		case 0xeffdd:
		case 0xeffde:
		case 0xeffdf:
			// mouse
			data = d_mouse->read_io8(addr);
			break;
		case 0xeffe0:
			// keyboard scan data
			data = d_key->read_io8(addr);
			break;
		case 0xeffe1:
			// kb type
			data = 1;
			break;
		case 0xeffe2:
			// tv super inpose (cannot read)
			break;
		case 0xeffe4:
			// psg data
			data = d_psg[0]->read_io8(1);
			break;
		case 0xeffe5:
			// psg reg (cannot read)
			break;
		case 0xeffe6:
			// ex psg data
			if (IOPORT_USE_EXPSG) {
				data = d_psg[1]->read_io8(1);
			}
			break;
		case 0xeffe7:
			// ex psg reg (cannot read)
			break;
		case 0xeffe8:
			// bank register
			break;
		case 0xeffe9:
			// IG mode register timing status
			data = REG_IGMODE;
			break;
		case 0xeffea:
			// IG en register
			break;
		case 0xeffeb:
			// system mode
			data = REG_SYS_MODE & 0x07;
			break;
	}

#ifdef _DEBUG_IO_COMMON
		if ((addr & 0xfff00) == 0xeff00) {
			uint8 ch = (data < 0x20 || data > 0x7f) ? 0x20 : data;
			emu->out_debugf("mr %04x=%02x %c",addr,data,ch);
		}
#endif

	// send trap signal when access the address 0xbfxxx 
	if (addr_seg04 == S1_ADDR_TRAP) {
		REG_TRAPF |= TRAPF_SIGNAL;
		d_board->write_signal(SIG_CPU_NMI, SIG_NMI_TRAP_MASK, SIG_NMI_TRAP_MASK);
	}

	return data;
}

void MEMORY::latch_address(uint32 addr, int *wait)
{
	addr = address_mapping(addr & 0xffff);
	addr_seg04 = (addr & 0xff000); // 4KB  block
	addr_comio = (addr & 0xeffff);

	// if access io port, cpu speed go down to 1MHz
	if (addr < 0xf0000) {
		// S1 mode
		addr_bank = (addr >> S1_BANK_SIZE);
		set_s1_cpu_wait(addr, addr_seg04, addr_bank, addr_comio, 0x03, 0x05, wait);
	} else {
		// L3 mode or upper 0xf0000
		addr_64kb = (addr & 0xffff);
		addr_bank = (addr_64kb >> L3_BANK_SIZE);
		set_l3_cpu_wait(addr, addr_bank, addr_comio, 0x03, 0x05, wait);
	}
}

void MEMORY::write_io8(uint32 addr, uint32 data)
{

}

void MEMORY::write_signal(int id, uint32 data, uint32 mask)
{
	switch(id) {
//		case SIG_MEMORY_PIA_PA:
//			// from PIA portA (L3)
//			mem_bank_reg1 = data & 0xce;
//			change_l3_memory_bank();
//			break;
		case SIG_CPU_BABS:
			if ((data & mask) == 0x01) {
				// interrupt
				// go to system mode
				if (NOW_USER_MODE) change_system_mode(false);
			}
			break;
		case SIG_CPU_RESET:
			now_reset = (data & mask) ? true : false;
			if (now_reset == true) {
				// now on reset signal

				// reset igmode
				set_igmode(0);
				ig_enreg = 0;
			} else {
				// turn off reset signal

				// set s1 or l3 mode
				REG_SYS_MODE = emu->get_parami(VM::ParamSysMode);
				reset_s1_l3_mode();
				// init memory map
				reset_l3_membank();
				// stop trace counter
				reset_trace_counter();
			}
			break;
	}
}

uint32 MEMORY::read_signal(int id)
{
	uint32 data = 0xff;

	return data;
}

void MEMORY::change_l3_memory_bank()
{
//	mem_bank_reg1;	// from 0xffc0
//	mem_bank_reg2;	// from 0xffe8
	uint8 *exram00 = NULL;
	uint8 *exram04 = NULL;
	uint8 *exram80 = NULL;

	// have extended ram
	if (exram_size >= 64 && exram != NULL) {
		exram00 = exram + 0x0000;
		exram04 = exram + 0x0400;
		exram80 = exram + 0x8000;
	}

	memset(l3wait, 0, sizeof(l3wait));

	// 0x0000 - 0x7fff
	if (mem_bank_reg2 & 0x08) {
		if (mem_bank_reg2 & 0x20) {
			// change to the extended ram both r/w
			L3_SET_BANK(0x0000, 0x7fff, exram00, exram00, 0x33);
		} else {
			// change to the extended ram only write
			L3_SET_BANK(0x0000, 0x7fff, exram00, ram + 0x0000, 0x37);
		}
	} else {
		// change to standard
		L3_SET_BANK(0x0000, 0x7fff, ram + 0x0000, ram + 0x0000, 0x77);
	}

	// 0x0400 - 0x43ff (vram area)
	// exchange vram area to/from extended ram
	if ((mem_bank_reg2 & 0x0c) == 0x04 || (mem_bank_reg2 & 0x0c) == 0x08) {
		// select extend ram
		if (mem_bank_reg2 & 0x20) {
			// read/write
			L3_SET_BANK(0x0400, 0x43ff, exram04, exram04, 0x33);
		} else {
			// write only
			L3_SET_BANK(0x0400, 0x43ff, exram04, ram + 0x0400, 0x37);
		}
	} else {
		// select standard ram
		L3_SET_BANK(0x0400, 0x43ff, ram + 0x0400, ram + 0x0400, 0x77);
	}

	// 0x8000 - 0xffff
	if ((mem_bank_reg2 & 0x50) == 0x50) {
		// change to the extended ram both r/w
		L3_SET_BANK(0x8000, 0xffff, exram80, exram80, 0x33);
	} else {
		//
		// decide the area to write
		if (mem_bank_reg2 & 0x10) {
			// change to the extended ram only write
			L3_SET_BANK(0x8000, 0xffff, exram80, NULL, 0x30);
		} else {
			// change to the standard
			L3_SET_BANK(0x8000, 0xffff, ram + 0x8000, NULL, 0x70);
		}

		//
		// decide the area to read
		L3_SET_BANK(0x8000, 0x9fff, NULL, ram + 0x8000, 0x07);
		L3_SET_BANK(0xa000, 0xffff, NULL, l3rom + 0x0000, 0x00);
		if (IOPORT_USE_3FDD && rom1805_loaded) {
			// for 3inch compact floppy
			L3_SET_BANK(0xf800, 0xffff, NULL, rom1805, 0x00);
		}
		if (IOPORT_USE_5FDD && rom1802_loaded) {
			// for 5.25inch mini floppy
			L3_SET_BANK(0xf800, 0xffff, NULL, rom1802, 0x00);
		}

		if (!(mem_bank_reg1 & 0x02)) {
			if (mem_bank_reg1 & 0x40) {
				// r/w to ram (this means cannot read rom data !)
				L3_SET_BANK(0xa000, 0xbfff, NULL, ram + 0xa000, 0x07);
			}
		}
		if (!(mem_bank_reg1 & 0x04)) {
			if (mem_bank_reg1 & 0x80) {
				// r/w to ram (this means cannot read rom data !)
				L3_SET_BANK(0xc000, 0xdfff, NULL, ram + 0xc000, 0x07);
			}
		}
		if (!(mem_bank_reg1 & 0x08)) {
			if (mem_bank_reg1 & 0x80) {
				// r/w to ram (this means cannot read rom data !)
				L3_SET_BANK(0xe000, 0xefff, NULL, ram + 0xe000, 0x07);
			}
		}

		if (mem_bank_reg2 & 0x01) {
			// read from ram (this means cannot read rom data !)
			L3_SET_BANK(0xf000, 0xfeff, NULL, ram + 0xf000, 0x07);
		} else {
			// read from rom , write to ram
			L3_SET_BANK(0xf000, 0xfeff, NULL, l3rom + 0x5000, 0x00);
			if (IOPORT_USE_3FDD && rom1805_loaded) {
				// for 3inch compact floppy
				L3_SET_BANK(0xf800, 0xfeff, NULL, rom1805, 0x00);
			}
			if (IOPORT_USE_5FDD && rom1802_loaded) {
				// for 5.25inch mini floppy
				L3_SET_BANK(0xf800, 0xfeff, NULL, rom1802, 0x00);
			}
		}
		if (mem_bank_reg2 & 0x02) {
			// r/w to ram (this means cannot read rom data !)
			L3_SET_BANK(0xfff0, 0xffff, NULL, ram + 0xfff0, 0x07);
		} else {
			// read from rom , write to ram
			L3_SET_BANK(0xfff0, 0xffff, NULL, l3rom + 0x5ff0, 0x00);
		}
	}
}

/// get L3 vram area 
uint8* MEMORY::get_l3vram() {
// vram is always the standard ram
//	return ((mem_bank_reg2 & 0x04) ? ram2 : ram );
	return ram;
}

/// get text vram
uint8* MEMORY::get_s1vtram() {
	return vtram;
}

/// get text color ram
uint8* MEMORY::get_s1vcram() {
	return vcram;
}

/// get graphic vram red (is main ram 0x0000 - 0x3fff)
uint8* MEMORY::get_s1vgrram() {
	return ram;
}

/// get graphic vram green
uint8* MEMORY::get_s1vggram() {
	return vgram;
}

/// get graphic vram blue
uint8* MEMORY::get_s1vgbram() {
	return &vgram[0x4000];
}

/// process trace counter if need
void MEMORY::process_trace_counter(int clk)
{
	if (trace_counter_phase != TRACE_COUNTER_PHASE_IDLE) {
		if (trace_counter_phase == TRACE_COUNTER_PHASE_WAITING) {
			// declease trace counter
			set_trace_counter(clk);
		} else if (trace_counter_phase == TRACE_COUNTER_PHASE_FIRED) {
			// trace counter reset
			clear_trace_counter(clk);
		}
	}
}

void MEMORY::fetch_trace_counter(int clk) {
	// no start trace counter if delay time is 0.
	if (FUSE_MODE_DELAY == 0) return;
//	pc = 0;
	trace_counter_phase = TRACE_COUNTER_PHASE_WAITING;
	// set trace count ($fe10)
	trace_count_remain = ((FUSE_MODE_DELAY + 1) << (NOW_2MHZ ? 0 : 1));
	prev_trace_clock = clk;
#ifdef _DEBUG_TRACE_COUNTER
	emu->out_debugf(_T("trace_fetch: clk:%d remain:%d"), clk, trace_count_remain);
#endif
}

void MEMORY::set_trace_counter(int clk) {
	int sum = clk - prev_trace_clock;
	if (sum < 0) sum += CLOCKS_CYCLE;
	trace_count_remain -= sum;
	prev_trace_clock = clk;
#ifdef _DEBUG_TRACE_COUNTER
	emu->out_debugf(_T("trace_set: clk:%d sum:%d remain:%d"), clk, sum, trace_count_remain);
#endif
	if (trace_count_remain <= 0) {
		fire_trace_counter(clk);
	}
}

void MEMORY::fire_trace_counter(int clk) {
	if (FUSE_TRACE_MASK) {
		// send NMI interrupt when trace NMI is not masked
		d_board->write_signal(SIG_CPU_NMI, SIG_NMI_TRACE_MASK, SIG_NMI_TRACE_MASK);
#ifdef _DEBUG_TRACE_COUNTER
		emu->out_debugf(_T("trace_fire: clk:%d remain:%d NMI on"), clk, trace_count_remain);
#endif
	} else {
		// go to user mode when trace NMI is masked
		if (NOW_SYSTEM_MODE) change_system_mode(true);
#ifdef _DEBUG_TRACE_COUNTER
		emu->out_debugf(_T("trace_fire: clk:%d remain:%d NMI masked"), clk, trace_count_remain);
#endif
	}
	trace_counter_phase = TRACE_COUNTER_PHASE_FIRED;
}

void MEMORY::clear_trace_counter(int clk)
{
	// wait until next opcode read
	int curclk = (current_clock() % CLOCKS_CYCLE);
	if (curclk < clk) return;

	int sum = clk - prev_trace_clock;
	if (sum < 0) sum += CLOCKS_CYCLE;
	if (sum > 0) {
		if (FUSE_TRACE_MASK) {
			// clear NMI
			d_board->write_signal(SIG_CPU_NMI, 0, SIG_NMI_TRACE_MASK);
#ifdef _DEBUG_TRACE_COUNTER
			emu->out_debugf(_T("trace NMI off: clk:%d sum:%d"), clk, sum);
#endif
		}
		trace_counter_phase = TRACE_COUNTER_PHASE_IDLE;
		trace_count_remain = 0;
		prev_trace_clock = 0;
#ifdef _DEBUG_TRACE_COUNTER
		emu->out_debugf(_T("trace_clear: clk:%d curclk:%d sum:%d"), clk, curclk, sum);
#endif
	}
}

/// memory mapping
uint32 MEMORY::address_mapping(uint32 addr)
{
	if (NOW_S1_MODE) {
		// S1 mode
		if (NOW_SYSTEM_MODE && (addr & 0xe000) == 0xe000) {
			// system mode
			addr = ((addr & 0x0fff) | ((addr & 0xf000) == 0xf000 ? 0xef000 : 0x84000));
		} else {
			// mapping
			addr = ((addr & 0x0fff) | (addr_map[REG_ADDRSEG][(addr & 0xf000) >> 12] << 12));
		}
	} else {
		// L3 mode
		addr |= 0xf0000;
	}
	return addr;
}

/// system <-> user mode
void MEMORY::change_system_mode(bool usermode)
{
	if (usermode && FUSE_MODE_DELAY) {
		// go to user mode
		REG_SYS_MODE &= ~SYS_MODE_SU;
		// all interrupt ok
		d_board->write_signal(SIG_FUSE_INTR_MASK, 0, 1);
	} else if (!usermode) {
		// go to system mode
		REG_SYS_MODE |= SYS_MODE_SU;
#ifdef _DEBUG
//		dasm.trace_on(50);
#endif
		// all interrupt mask except trace processing.
		if (trace_counter_phase == TRACE_COUNTER_PHASE_IDLE) {
			d_board->write_signal(SIG_FUSE_INTR_MASK, 1, 1);
		}
	}
#ifdef _DEBUG_TRACE_COUNTER
	emu->out_debugf(_T("change_system_mode: %s trace_phase:%d"), (NOW_SYSTEM_MODE ? _T("S") : _T("U")), trace_counter_phase);
#endif
}

/// set S1/L3 mode
void MEMORY::reset_s1_l3_mode()
{
	REG_SYS_MODE |= SYS_MODE_SU;
	if (NOW_S1_MODE) {
		REG_SYS_MODE |= SYS_MODE_CLK;
	} else {
		REG_SYS_MODE &= ~SYS_MODE_CLK;
	}
	set_cpu_clock(REG_SYS_MODE);

	// initialize crtc clock 
	d_crtc->write_signal(SIG_HD46505_CHR_CLOCKS, CLOCKS_1MHZ * (NOW_S1_MODE || (REG_MODE_SEL & 0x80) ? 2 : 1), 0xffffffff);
	d_disp->write_signal(SIG_DISPLAY_CHR_CLOCKS, CLOCKS_1MHZ * (NOW_S1_MODE || (REG_MODE_SEL & 0x80) ? 2 : 1), 0xffffffff);
}

/// set cpu clock (bit1)
void MEMORY::set_cpu_clock(uint8 data)
{
	REG_SYS_MODE = ((data & SYS_MODE_CLK) ? (REG_SYS_MODE | SYS_MODE_CLK) : (REG_SYS_MODE & ~SYS_MODE_CLK));
	d_event->write_signal(SIG_EVENT_CPU_HALF_SPEED, NOW_2MHZ ? 0 : 1, 1);
}

/// (re)allocate extended ram 
void MEMORY::set_extended_ram()
{
	// 0:0KB 1:64KB 2:128KB 3:256KB 4:512KB
	int size = (config.exram_size_num ? ((1 << config.exram_size_num) * 32) : 0);
	if (exram != NULL) {
		delete [] exram;
		exram_size = 0;
		exram = NULL;

		// reset ext ram 0x00000 - 0x7ffff
		S1_SET_BANK(0x00000, 0x7ffff, wdmy, rdmy, 0);
	}
	if (size > 0) {
		exram = new uint8[size * 1024];
		exram_size = size;

		memset(exram, 0, sizeof(uint8)* size * 1024);

		// ext ram 0x00000 - 0x0ffff - 0x7ffff
		S1_SET_BANK(0x00000, size*1024-1, exram, exram, 0x33);
	}
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

#if 0
void MEMORY::event_callback(int event_id, int err)
{
	if (event_id == SIG_TRACE_COUNTER) {
		// event stop
//		cancel_event(trace_counter_id);
		trace_counter_id = -1;

		pc = d_cpu->get_pc();

#if 0
		if (pc == pc_prev || pc_prev == TRACE_COUNTER_PHASE_WAITING) {
			// inhibit the NMI interrupt until next opcode is executed.
			pc_prev = pc;
			// next event
			register_event(this, SIG_TRACE_COUNTER, 2, false, &trace_counter_id);
			emu->out_debugf(_T("trace_fire: clk:%lld pc:%04x next"), current_clock() % CLOCKS_CYCLE, pc);
		} else
#endif
		{
			// go to user mode
			if (NOW_SYSTEM_MODE) change_system_mode(true);
			// send NMI interrupt
			if (FUSE_TRACE_MASK) {
				d_board->write_signal(SIG_CPU_NMI, SIG_NMI_TRACE_MASK, SIG_NMI_TRACE_MASK);
//				emu->out_debugf(_T("trace_fire: clk:%lld pc:%04x NMI on"), current_clock() % CLOCKS_CYCLE, pc);
//			} else {
//				emu->out_debugf(_T("trace_fire: clk:%lld pc:%04x NMI masked"), current_clock() % CLOCKS_CYCLE, pc);
			}
			pc_prev = TRACE_COUNTER_PHASE_FIRED;
		}
	}
}
#endif

// ----------------------------------------------------------------------------

void MEMORY::update_config()
{
}

// ----------------------------------------------------------------------------

void MEMORY::save_state(FILEIO *fio)
{
	struct vm_state_st_s1 vm_state;

	//
	vm_state_ident.version = 0x43;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state) + (exram_size * 1024);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	memcpy(vm_state.ram,  ram,  sizeof(ram));
	memcpy(vm_state.vgram, vgram, sizeof(vgram));
	memcpy(vm_state.vtram, vtram, sizeof(vtram));
	memcpy(vm_state.vcram, vcram, sizeof(vcram));
	memcpy(vm_state.ig_ram, ig_ram, sizeof(ig_ram));

//	vm_state.trace_counter_id = trace_counter_id;
//	vm_state.pc = pc;
//	vm_state.pc_prev = pc_prev;
	vm_state.trace_counter_phase = trace_counter_phase;
	vm_state.trace_count_remain = trace_count_remain;
	vm_state.prev_trace_clock = prev_trace_clock;

	vm_state.mem_bank_reg1 = mem_bank_reg1;
	vm_state.mem_bank_reg2 = mem_bank_reg2;
	vm_state.mem_bank_mask = mem_bank_mask ? 1 : 0;

	// save config
	vm_state.dipswitch = config.dipswitch;
	vm_state.sys_mode = REG_SYS_MODE;
	vm_state.fdd_type = config.fdd_type;
	vm_state.io_port = config.io_port;
	vm_state.mode_sel = REG_MODE_SEL;
	vm_state.color_reg = color_reg;
	vm_state.ig_enreg = ig_enreg;
	vm_state.reg_fuse = REG_FUSE;
	vm_state.reg_trapf = REG_TRAPF;
	vm_state.reg_busctrl = REG_BUSCTRL;

	// version 0x43
	memcpy(vm_state.v43.addr_map, addr_map, sizeof(addr_map));
	vm_state.v43.reg_addrseg = REG_ADDRSEG;

	vm_state.v43.exram_size = exram_size;
	
	fio->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fio->Fwrite(&vm_state, sizeof(vm_state), 1);

	// save extended ram
	if (exram_size > 0) {
		fio->Fwrite(exram, exram_size * 1024, 1);
	}
}

void MEMORY::load_state(FILEIO *fio)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st_l3 vm_state_l3;
	struct vm_state_st_s1 vm_state_s1;

	FIND_STATE_CHUNK(fio, vm_state_i);
	if (vm_state_i.version < 0x41) {
		// L3
		READ_STATE_CHUNK(fio, vm_state_i, vm_state_l3);

		// set B mode
		memset(addr_map, 0, sizeof(addr_map));
		REG_SYS_MODE = SYS_MODE_SU;

		// copy values
		memcpy(ram,  vm_state_l3.ram,  sizeof(ram));
		memcpy(vgram, vm_state_l3.color_ram, sizeof(vm_state_l3.color_ram));
		memcpy(ig_ram, vm_state_l3.ig_ram, sizeof(ig_ram));
		
		mem_bank_reg1 = vm_state_l3.mem_bank_reg1;
		mem_bank_reg2 = vm_state_l3.mem_bank_reg2;
		mem_bank_mask = false;
	
//		trace_counter_id = vm_state_l3.trace_counter_id;
//		pc = vm_state_l3.pc;
//		pc_prev = vm_state_l3.pc_prev;
		trace_counter_phase = TRACE_COUNTER_PHASE_IDLE;
		trace_count_remain = 0;
		prev_trace_clock = 0;

		// calc values
		// rs bit on mode_sel is set in COMM device. 
		REG_MODE_SEL = ram[0xffd0];
		d_disp->write_io8(0xffd0, REG_MODE_SEL);
		reset_s1_l3_mode();

		d_disp->write_io8(0xffd6, ram[0xffd6] & 0x08);	// interace sel
		color_reg = ram[0xffd8];	// color reg
	
//		write_data8(0xffea, ram[0xffea]);	// IG en register
		ig_enreg = ram[0xffea] & 0x07;
	
		// load config
		config.dipswitch = vm_state_l3.dipswitch;
		config.fdd_type = vm_state_l3.fdd_type;
		config.io_port = vm_state_l3.io_port;

		// load extended memory
		exram_size = 64;	// 64KB constant
		config.exram_size_num = 1;
		set_extended_ram();
		memcpy(exram, vm_state_l3.ram2, sizeof(vm_state_l3.ram2));

	} else {
		// S1
		READ_STATE_VARIABLE_CHUNK(fio, vm_state_i, vm_state_s1);
		// copy values
		memcpy(ram,  vm_state_s1.ram,  sizeof(ram));
		memcpy(vgram, vm_state_s1.vgram, sizeof(vgram));
		memcpy(vtram, vm_state_s1.vtram, sizeof(vtram));
		memcpy(vcram, vm_state_s1.vcram, sizeof(vcram));
		memcpy(ig_ram, vm_state_s1.ig_ram, sizeof(ig_ram));

//		trace_counter_id = vm_state_s1.trace_counter_id;
//		pc = vm_state_s1.pc;
//		pc_prev = vm_state_s1.pc_prev;
		if (vm_state_i.version >= 0x42) {
			trace_counter_phase = vm_state_s1.trace_counter_phase;
			trace_count_remain = vm_state_s1.trace_count_remain;
			prev_trace_clock = vm_state_s1.prev_trace_clock;
		} else {
			trace_counter_phase = TRACE_COUNTER_PHASE_IDLE;
			trace_count_remain = 0;
			prev_trace_clock = 0;
		}

		mem_bank_reg1 = vm_state_s1.mem_bank_reg1;
		mem_bank_reg2 = vm_state_s1.mem_bank_reg2;
		mem_bank_mask = (vm_state_s1.mem_bank_mask != 0);

		config.dipswitch = vm_state_s1.dipswitch;
		REG_SYS_MODE = vm_state_s1.sys_mode;
		set_cpu_clock(REG_SYS_MODE);

		config.fdd_type = vm_state_s1.fdd_type;
		config.io_port = vm_state_s1.io_port;

		REG_MODE_SEL = vm_state_s1.mode_sel;	// mode sel
		color_reg = vm_state_s1.color_reg;	// color reg
		ig_enreg = vm_state_s1.ig_enreg;	// IG en register
		REG_FUSE = vm_state_s1.reg_fuse;
		REG_TRAPF = vm_state_s1.reg_trapf;
		REG_BUSCTRL = vm_state_s1.reg_busctrl;

		if (vm_state_i.version >= 0x43) {
			// version 0x43
			memcpy(addr_map, vm_state_s1.v43.addr_map, sizeof(addr_map));
			REG_ADDRSEG = vm_state_s1.v43.reg_addrseg;
			// load extended memory
			exram_size = vm_state_s1.v43.exram_size;
		} else {
			// version ..0x42
			memcpy(addr_map[0], vm_state_s1.v42.addr_map, sizeof(addr_map[0]));
			// load extended memory
			exram_size = vm_state_s1.v42.exram_size;
			// shift color ram if L3 mode
			if (NOW_L3_MODE) {
				for(int i=0; i<L3_VRAM_SIZE; i++) {
					vgram[i] = vgram[i + L3_ADDR_VRAM_START];
				}
				for(int i=L3_VRAM_SIZE; i<L3_ADDR_VRAM_END; i++) {
					vgram[i] = 0;
				}
			}
			// back position
			fio->Fseek((long)(sizeof(vm_state_s1.v42)-sizeof(vm_state_s1.v43)), FILEIO_SEEK_CUR);
		}

		if (exram_size > 512) exram_size = 64;
		config.exram_size_num = 0;
		for(int i=1; i<=4; i++) {
			if (exram_size == (1 << i) * 32) {
				config.exram_size_num = i;
				break;
			}
		}
		set_extended_ram();

		if (exram_size > 0) {
			fio->Fread(exram, exram_size * 1024, 1);
		}

		write_data8(0xffd0, REG_MODE_SEL);	// mode sel
	}

	change_l3_memory_bank();
}
