/** @file memory.h

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ memory ]
*/

#ifndef _MEMORY_H_
#define _MEMORY_H_

#include "../vm.h"
#include "../../emu.h"
#include "../device.h"
#include "../../config.h"
#include "../hd46505.h"
#include "sound.h"
#include "display.h"
#include "comm.h"
#include "cmt.h"
#include "pia.h"
#include "via.h"
#include "../event.h"

#define SIG_MEMORY_DISP		0
#define SIG_MEMORY_VSYNC	1
//#define SIG_MEMORY_PIA_PA	2
#define SIG_TRACE_COUNTER	4
#define SIG_TRACE_COUNTER_STOP	5

#define S1_BANK_SIZE	11	///< 2Kbytes
#define L3_BANK_SIZE	4	///< 16bytes


/**
	@brief Memory access
*/
class MEMORY : public DEVICE
{
private:
	DEVICE *d_cpu, *d_pia, *d_crtc, *d_disp, *d_acia;
	DEVICE *d_key, *d_mouse, *d_sound, *d_comm, *d_cmt, *d_timer;
	DEVICE *d_kanji, *d_fdc3, *d_fdc5, *d_fdc5_ex, *d_fdd;
	DEVICE *d_psg[2], *d_psg3, *d_psg9;
	DEVICE *d_pia_ex, *d_acia_ex, *d_comm1, *d_board;
	DEVICE *d_event;

	uint8 l3rom[0x6000];		///< 24KB
	uint8 s1rom[0x10000];		///< 64KB
	uint8 s1dicrom[0x8000];		///< 32KB
	uint8 rom1802[0x800];
	uint8 rom1805[0x800];


	uint8 ram[0x10000];			///< main ram (S1 is also used as graphic vram red)
//	uint8 color_ram[0x4000];	///< 0x0400 - 0x43ff -> use vgram
	uint8 ig_ram[0x0800 * 3];	///< 0xa000 - 0xa7ff * 3 (b,r,g)
//	uint8 ram2[0x10000];		///< extended ram
	int   exram_size;			///< extended ram size (KB)
	uint8 *exram;				///< extended ram (variable)

	uint8 vgram[0x8000];		///< S1 graphic ram (green,blue)(32KB)
	uint8 vtram[0x800];			///< S1 text ram (2KB)
	uint8 vcram[0x800];			///< S1 color ram (2KB)
	uint8 *font;				///< font rom (4KB)

	uint8 addr_map[64][16];		///< address mapping table

	uint8 wdmy[1 << S1_BANK_SIZE];
	uint8 rdmy[1 << S1_BANK_SIZE];
	uint8* l3wbank[0x10000 >> L3_BANK_SIZE];
	uint8* l3rbank[0x10000 >> L3_BANK_SIZE];
	uint8* s1wbank[0x100000 >> S1_BANK_SIZE];
	uint8* s1rbank[0x100000 >> S1_BANK_SIZE];

	/// bit0:read wait bit4:write wait
	/// bit1,5: refresh memory wait
	/// bit2,6: disptmg wait
	uint8  l3wait[0x10000 >> L3_BANK_SIZE];
	uint8  s1wait[0x100000 >> S1_BANK_SIZE];

	uint8 color_reg;		///< 0xffd8

//	uint8 ig_enable;		///< 0xffe9
	uint8 ig_enreg;			///< 0xffea

	uint8 mem_bank_reg1;	///< 0xffc0
	uint8 mem_bank_reg2;	///< 0xffe8
	bool  mem_bank_mask;	///< 0xffc0 mask

//	uint8 vram_sel;
//	uint8 cpu_wait;			///< io access (bit1) 0:1MHz 2:2MHz

//	int   trace_counter_id;
//	uint32 pc, pc_prev;
	int   trace_counter_phase;
	int   trace_count_remain;
	int   prev_trace_clock;

	bool  l3rom_loaded;
	bool  s1rom1_loaded;
	bool  s1rom2_loaded;
	bool  s1dicrom_loaded;
	bool  rom1802_loaded;
	bool  rom1805_loaded;
	bool  rom_loaded_at_first;

	uint32 addr_bank;
	uint32 addr_seg04, addr_seg16, addr_seg32;
	uint32 addr_comio, addr_64kb;

#ifdef USE_RTC
	uint8 rtc_reg_sel;
	uint8 rtc_buf;
#endif

	//for resume
#pragma pack(1)
	struct vm_state_st_l3 {
		uint8 ram[0x10000];
		uint8 ram2[0x10000];		// extended ram
		uint8 color_ram[0x4000];	// 0x0400 - 0x43ff
		uint8 ig_ram[0x0800 * 3];	// 0xa000 - 0xa7ff * 3 (b,r,g)

		uint8 mem_bank_reg1;	// from PIA port A
		uint8 mem_bank_reg2;	// from 0xffe8

		int   trace_counter_id;
		uint32 pc;
		uint32 pc_prev;

		// from config
		uint8 dipswitch;
		int fdd_type;
		int io_port;

		char  reserved[9];
	};

	struct vm_state_st_s1 {
		uint8 ram[0x10000];
		uint8 vgram[0x8000];		// graphic ram
		uint8 vtram[0x800];			// text chr
		uint8 vcram[0x800];			// text color
		uint8 ig_ram[0x0800 * 3];	// ig (b,r,g)

//		int   trace_counter_id;
//		uint32 pc;
//		uint32 pc_prev;
		int   trace_counter_phase;	// version 0x42
		int   trace_count_remain;	// version 0x42
		int   prev_trace_clock;		// version 0x42

		uint8 mem_bank_reg1;	// from 0xffc0
		uint8 mem_bank_reg2;	// from 0xffe8
		uint8 mem_bank_mask;	// from 0xffc0 mask
		char  reserved1;

		// from config
		uint8 dipswitch;
		uint8 sys_mode;
		int fdd_type;
		int io_port;

		uint8 mode_sel;
		uint8 color_reg;
		uint8 ig_enreg;
		uint8 reg_fuse;
		uint8 reg_trapf;
		uint8 reg_busctrl;

		union {
			struct {
				uint8 addr_map[16];
				char  reserved2[256];	// for OS-9 exboard

				char  exram_ident[14];
				uint16 exram_size;
				// exram (variable)
				// uint8 exram[0 - 512KB]
			} v42;
			struct {
				uint8 addr_map[64][16];
				uint8 reg_addrseg;

				char reserved3[7];
				char exram_ident[6];
				uint16 exram_size;
				// exram (variable)
				// uint8 exram[0 - 512KB]
			} v43;
		};
	};
#pragma pack()

	void SET_BANK(int s, int e, uint8* w, uint8* r, uint8** wbank, uint8** rbank, int bank_size, uint8 *waitbank, uint8 waitval);

	void load_rom_files();
	void reset_sram(uint8 *sram, size_t size);
	void reset_igram();
	void reset_s1_membank();
	void reset_l3_membank();
	void reset_trace_counter();
	void set_igmode(uint32 data);
	inline void process_trace_counter(int clk);
	inline void fetch_trace_counter(int clk);
	inline void set_trace_counter(int clk);
	inline void fire_trace_counter(int clk);
	inline void clear_trace_counter(int clk);

	void change_l3_memory_bank();
	void change_system_mode(bool usermode);
	void reset_s1_l3_mode();
	inline void set_cpu_clock(uint8 data);

	void set_extended_ram();

public:
	MEMORY(VM* parent_vm, EMU* parent_emu, const char* identifier) : DEVICE(parent_vm, parent_emu, identifier)
	{
		set_class_name("MEMORY");
	}
	~MEMORY() {}

	// common functions
	void initialize();
	void reset();
	void release();
	void write_data8w(uint32 addr, uint32 data, int *wait);
	uint32 read_data8w(uint32 addr, int *wait);
	void latch_address(uint32 addr, int *wait);

//	void write_data16(uint32 addr, uint32 data) {
//		write_data8(addr, data & 0xff); write_data8(addr + 1, data >> 8);
//	}
//	uint32 read_data16(uint32 addr) {
//		return read_data8(addr) | (read_data8(addr + 1) << 8);
//	}
	void write_io8(uint32 addr, uint32 data);
	void write_signal(int id, uint32 data, uint32 mask);
	uint32 read_signal(int id);

	// unique function
	void set_context_cpu(DEVICE* device) {
		d_cpu = device;
	}
	void set_context_pia(DEVICE* device) {
		d_pia = device;
	}
	void set_context_crtc(DEVICE* device) {
		d_crtc = device;
	}
	void set_context_display(DEVICE* device) {
		d_disp = device;
	}
	void set_context_acia(DEVICE* device) {
		d_acia = device;
	}
	void set_context_key(DEVICE* device) {
		d_key = device;
	}
	void set_context_mouse(DEVICE* device) {
		d_mouse = device;
	}
	void set_context_sound(DEVICE* device) {
		d_sound = device;
	}
	void set_context_psg(DEVICE* device0, DEVICE* device1) {
		d_psg[0] = device0;
		d_psg[1] = device1;
	}
	void set_context_psg3(DEVICE* device) {
		d_psg3 = device;
	}
	void set_context_psg9(DEVICE* device) {
		d_psg9 = device;
	}
	void set_context_comm(DEVICE* device) {
		d_comm = device;
	}
	void set_context_cmt(DEVICE* device) {
		d_cmt = device;
	}
	void set_context_timer(DEVICE* device) {
		d_timer = device;
	}
	void set_context_kanji(DEVICE* device) {
		d_kanji = device;
	}
	void set_context_fdc(DEVICE* device3, DEVICE* device5, DEVICE* device5e) {
		d_fdc3 = device3;
		d_fdc5 = device5;
		d_fdc5_ex = device5e;
	}
	void set_context_fdd(DEVICE* device) {
		d_fdd = device;
	}
	void set_context_pia_ex(DEVICE* device) {
		d_pia_ex = device;
	}
	void set_context_acia_ex(DEVICE* device) {
		d_acia_ex = device;
	}
	void set_context_comm1(DEVICE* device) {
		d_comm1 = device;
	}
	void set_context_board(DEVICE* device) {
		d_board = device;
	}
	void set_context_event(DEVICE* device) {
		d_event = device;
	}
	uint8* get_l3vram();
	uint8* get_s1vtram();
	uint8* get_s1vcram();
	uint8* get_s1vgrram();
	uint8* get_s1vggram();
	uint8* get_s1vgbram();

	uint8* get_color_ram() {
		return vgram;
	}
	uint8* get_ig_ram() {
		return ig_ram;
	}
	void set_font(uint8 *val) {
		font = val;
	}

	uint32 address_mapping(uint32 addr);

//	void event_callback(int event_id, int err);
	void update_config();

	void save_state(FILEIO *fio);
	void load_state(FILEIO *fio);

};

#endif
