/** @file timer.cpp

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ timer ]
*/

#include "timer.h"

void TIMER::initialize()
{
	timer_irq = 0;
	time_mask = 0;
	cnt_2msec = 0;
	register_id = -1;

	register_vline_event(this);	// 60Hz timer
}

void TIMER::reset()
{
	timer_irq = 0;
	time_mask = 0;
	cnt_2msec = 0;

	if (register_id >= 0) {
		cancel_event(register_id);
		register_id = -1;
	}
	if (NOW_S1_MODE) {
		time_mask = 0x82;
		register_event(this, SIG_TIMER_1MSEC, 1000, true, &register_id);
	}
}

void TIMER::write_signal(int id, uint32 data, uint32 mask)
{
	switch (id) {
//		case SIG_TIMER_VSYNC:
//			if (data & mask) update_timer_clock();
//			break;
		case SIG_CPU_RESET:
			now_reset = (data & mask) ? true : false;
			if (!now_reset) {
				reset();
			}
			break;
	}
}

void TIMER::write_io8(uint32 addr, uint32 data)
{
	switch (addr & 0xffff) {
		case ADDR_TIMER_IRQ:
			break;
		case ADDR_TIME_MASK:
			time_mask = data & 0x83;
			if (time_mask & 0x80) {
				timer_irq &= 0x7f;
				// Clear FIRQ
				d_board->write_signal(SIG_CPU_FIRQ, 0, SIG_FIRQ_TIMER1_MASK);
			}
			if (time_mask & 0x02) {
				timer_irq &= 0xbf;
				// Clear FIRQ
				d_board->write_signal(SIG_CPU_FIRQ, 0, SIG_FIRQ_TIMER2_MASK);
			}
//			emu->out_debugf(_T("timer w %04x %02x"), addr, data);
			break;
	}
}

uint32 TIMER::read_io8(uint32 addr)
{
	uint32 data = 0;
	switch (addr & 0xffff) {
		case ADDR_TIMER_IRQ:
			data = timer_irq;
			timer_irq = 0;
			// Clear FIRQ
			d_board->write_signal(SIG_CPU_FIRQ, 0, SIG_FIRQ_TIMER_MASK);
//			emu->out_debugf(_T("timer r %04x %02x"), addr, data);
			break;
		case ADDR_TIME_MASK:
			break;
	}
	return data;
}

/// 60Hz timer
void TIMER::update_timer_clock()
{
	if (!(time_mask & 0x80) && !now_reset) {
		timer_irq |= 0x80;		// timer irq set
		// FIRQ interrupt
		d_board->write_signal(SIG_CPU_FIRQ, SIG_FIRQ_TIMER1_MASK, SIG_FIRQ_TIMER1_MASK);
//		emu->out_debugf("timer normal set irq %02x", timer_irq);
	}
}

/// 1msec fast timer
void TIMER::update_fast_timer_clock()
{
	if (((time_mask & 0x03) == 1 || ((time_mask & 0x03) == 0 && cnt_2msec != 0)) && !now_reset) {
		timer_irq |= 0x40;		// timer irq set
		// FIRQ interrupt
		d_board->write_signal(SIG_CPU_FIRQ, SIG_FIRQ_TIMER2_MASK, SIG_FIRQ_TIMER2_MASK);
//		emu->out_debugf("timer fast   set irq %02x", timer_irq);
	}
	cnt_2msec = (1-cnt_2msec);
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

void TIMER::event_vline(int v, int clock)
{
	if (v == LINES_PER_FRAME - 2) update_timer_clock();
}

void TIMER::event_callback(int event_id, int err)
{
	update_fast_timer_clock();
}

// ----------------------------------------------------------------------------

void TIMER::save_state(FILEIO *fio)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 0x41;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	vm_state.timer_irq = timer_irq;
	vm_state.time_mask = time_mask;
	vm_state.cnt_2msec = cnt_2msec;
	vm_state.register_id = register_id;

	fio->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fio->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void TIMER::load_state(FILEIO *fio)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;

	READ_STATE_CHUNK(fio, vm_state_i, vm_state);

	timer_irq = vm_state.timer_irq;
	time_mask = vm_state.time_mask;

	if (vm_state_i.version >= 0x41) {
		cnt_2msec = vm_state.cnt_2msec;
		register_id = vm_state.register_id;
	}
}
