/** @file via.h

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2012.06.08 -

	@brief [ via modoki (mcs6522) ]
*/

#ifndef _VIA_MODOKI_H_
#define _VIA_MODOKI_H_

#include "../vm.h"
#include "../../emu.h"
#include "../device.h"

#define SIG_VIA_PA	1
#define SIG_VIA_CA1	2
#define SIG_VIA_CA2	3
#define SIG_VIA_PB	4
#define SIG_VIA_CB1	5
#define SIG_VIA_CB2	6
#define SIG_VIA_CLOCK_UNIT 7

/**
	@brief via modoki (mcs6522) - Versatile Interface Adapter
*/
class VIA : public DEVICE
{
private:
	uint8	drb;	// Data Register B (ORB/IRB)
	uint8	dra;	// Data Register A (ORA/IRA)
	uint8	ddrb;	// Data Direction Register B
	uint8	ddra;	// Data Direction Register A

	int		t1l;	// Timer 1 latches
	int		t1c;	// Timer 1 counter
	int		t2l;	// Timer 2 latches
	int		t2c;	// Timer 2 counter

	uint8	sr;		// Shift Register
	uint8	acr;	// Auxiliary Control Register
	uint8	pcr;	// Peripheral Control Register
	uint8	ifr;	// Interrupt Flag Register
	uint8	ier;	// Interrupt Enable Register

	uint8	cra;	// Control Register A
	uint8	crb;	// Control Register B

	uint8	ca1;	// CA1 signal
	uint8	ca2;	// CA2 signal
	uint8	cb1;	// CB1 signal
	uint8	cb2;	// CB2 signal

	uint8	drb_res;	// IRB reserve data
	uint8	dra_res;	// IRA reserve data

	uint8	t1_pb7;	// PB7 with timer 1
	uint8	t2_pb6;	// PB6 with timer 2

	uint64	timer1_clock;	// timer1 start clock
	uint64	timer2_clock;	// timer2 start clock
	int		shift_timer;	// shift timer
	int		shift_count;	// shift count

	int ca2_register_id;
	int cb2_register_id;
	int t1_register_id;
	int t2_register_id;
	int s2_register_id;

	outputs_t outputs_pa;
	outputs_t outputs_ca2;
	outputs_t outputs_pb;
	outputs_t outputs_cb1;
	outputs_t outputs_cb2;
	outputs_t outputs_irq;

	uint8 now_irq;

	uint8 clk_unit;

	//for resume
#pragma pack(1)
	struct vm_state_st {
		int ca2_register_id;
		int cb2_register_id;
		int t1_register_id;
		int t2_register_id;

		int s2_register_id;
		uint8	drb;
		uint8	dra;
		uint8	ddrb;
		uint8	ddra;
		int		t1l;
		int		t1c;

		int		t2l;
		int		t2c;
		uint8	sr;
		uint8	acr;
		uint8	pcr;
		uint8	ifr;
		uint8	ier;
		uint8	cra;
		uint8	crb;
		uint8	ca1;

		uint8	ca2;
		uint8	cb1;
		uint8	cb2;
		uint8	drb_res;
		uint8	dra_res;
		uint8	t1_pb7;
		uint8	t2_pb6;

		uint8   now_irq;

//		uint32	timer1_clock;
//		uint32	timer2_clock;
//		int		shift_count;

		union {
			struct {
				uint32	timer1_clock;
				uint32	timer2_clock;
				int		shift_count;
				char reserved[8];
			} v1;
			struct {
				uint64	timer1_clock;
				uint64	timer2_clock;
				int		shift_count;
			} v2;
		} d;

		char reserved[4];
	};
#pragma pack()

	void set_ca2(uint8 val);
	void set_cb1(uint8 val);
	void set_cb2(uint8 val);
	void set_irq(uint8 data, uint8 mask);
	void calc_clock(int &tc, uint64 &clock);

	void shift_low();
	void shift();
	void set_shift_intr();
	void cancel_my_events();

public:
	VIA(VM* parent_vm, EMU* parent_emu, const char* identifier) : DEVICE(parent_vm, parent_emu, identifier) {
		set_class_name("VIA");
		init_output_signals(&outputs_pa);
		init_output_signals(&outputs_ca2);
		init_output_signals(&outputs_pb);
		init_output_signals(&outputs_cb1);
		init_output_signals(&outputs_cb2);
		init_output_signals(&outputs_irq);
	
		clk_unit = 1;
	}
	~VIA() {}

	// common functions
	void initialize();
	void reset();

	void write_io8(uint32 addr, uint32 data);
	uint32 read_io8(uint32 addr);

	void write_signal(int id, uint32 data, uint32 mask);

	// unique function
	void set_context_pa(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_pa, device, id, mask);
	}
	void set_context_ca2(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_ca2, device, id, mask);
	}
	void set_context_pb(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_pb, device, id, mask);
	}
	void set_context_cb1(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_cb1, device, id, mask);
	}
	void set_context_cb2(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_cb2, device, id, mask);
	}
	void set_context_irq(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_irq, device, id, mask);
	}

	// event callback
	void event_frame();
	void event_callback(int event_id, int err);

	void save_state(FILEIO *fp);
	void load_state(FILEIO *fp);
};

#endif
