/** @file win32_sound.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@brief [ win32 sound ]
*/

#include "emu.h"
#include "vm/vm.h"
#include "fileio.h"

#undef USE_SOUND_NOTIFY

#define DSOUND_SAMPLE (sound_samples)
#define DSOUND_BUFFER_SIZE (DWORD)(DSOUND_SAMPLE << 3)
#define DSOUND_BUFFER_HALF (DWORD)(DSOUND_SAMPLE << 2)

#define SOUND_RECORD_IN_MIXER

void EMU::EMU_SOUND()
{
	sound_rate = 0;
	sound_samples = 0;
	sound_latency_half = 0;
	sound_ok = sound_started = sound_finished = now_mute = now_rec_snd = false;
	sound_prev_time = 0;
	sound_nt_event = NULL;
	sound_threadid = 0;
	sound_thread = NULL;
	lpdsnt = NULL;
	lpds = NULL;
	lpdsp = NULL;
	ZeroMemory(&sound_nt_cs, sizeof(sound_nt_cs));
}

void EMU::initialize_sound(int rate, int samples, int latency)
{
	sound_rate = rate;
	sound_samples = samples;
	sound_latency_half = latency / 2;
	vm->initialize_sound(sound_rate, sound_samples);

	// initialize direct sound
	PCMWAVEFORMAT pcmwf;
	DSBUFFERDESC dsbd;
	WAVEFORMATEX wfex;

	if(FAILED(DirectSoundCreate(NULL, &lpds, NULL))) {
		return;
	}
	if(FAILED(lpds->SetCooperativeLevel(main_window_handle, DSSCL_PRIORITY))) {
		return;
	}

	// primary buffer
	ZeroMemory(&dsbd, sizeof(dsbd));
	dsbd.dwSize = sizeof(dsbd);
	dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;
	if(FAILED(lpds->CreateSoundBuffer(&dsbd, &lpdsp, NULL))) {
		return;
	}
	ZeroMemory(&wfex, sizeof(wfex));
	wfex.wFormatTag = WAVE_FORMAT_PCM;
	wfex.nChannels = 2;
#ifdef USE_AUDIO_U8
	wfex.wBitsPerSample = 8;
#else
	wfex.wBitsPerSample = 16;
#endif
	wfex.nSamplesPerSec = sound_rate;
	wfex.nBlockAlign = wfex.nChannels * wfex.wBitsPerSample / 8;
	wfex.nAvgBytesPerSec = wfex.nSamplesPerSec * wfex.nBlockAlign;
	if(FAILED(lpdsp->SetFormat(&wfex))) {
		return;
	}

	// secondary buffer
	ZeroMemory(&pcmwf, sizeof(pcmwf));
	pcmwf.wf.wFormatTag = WAVE_FORMAT_PCM;
	pcmwf.wf.nChannels = 2;
#ifdef USE_AUDIO_U8
	pcmwf.wBitsPerSample = 8;
#else
	pcmwf.wBitsPerSample = 16;
#endif
	pcmwf.wf.nSamplesPerSec = sound_rate;
	pcmwf.wf.nBlockAlign = pcmwf.wf.nChannels * pcmwf.wBitsPerSample / 8;
	pcmwf.wf.nAvgBytesPerSec = pcmwf.wf.nSamplesPerSec * pcmwf.wf.nBlockAlign;
	ZeroMemory(&dsbd, sizeof(dsbd));
	dsbd.dwSize = sizeof(dsbd);
	dsbd.dwFlags = DSBCAPS_STICKYFOCUS | DSBCAPS_GETCURRENTPOSITION2;
#ifdef USE_SOUND_NOTIFY
	dsbd.dwFlags |= DSBCAPS_CTRLPOSITIONNOTIFY;
#endif
	dsbd.dwBufferBytes = DSOUND_BUFFER_SIZE;
	dsbd.lpwfxFormat = (LPWAVEFORMATEX)&pcmwf;
	if(FAILED(lpds->CreateSoundBuffer(&dsbd, &lpdsb, NULL))) {
		return;
	}

#ifdef USE_SOUND_NOTIFY
	// create sound notify evnet
	if (FAILED(lpdsb->QueryInterface(IID_IDirectSoundNotify, (VOID**)&lpdsnt))) {
		out_logf(LOG_ERROR,_T("EMU::initialize_sound: QueryInterface %ld"), GetLastError());
		return;
	}
	if ((sound_nt_event = CreateEvent(NULL, FALSE, FALSE, _T("emu_sound_notify"))) == NULL) {
		out_logf(LOG_ERROR,_T("EMU::initialize_sound: CreateEvent %ld"), GetLastError());
		return;
	}

	// set fire point on sound buffer
	DSBPOSITIONNOTIFY dsnt[2];

	dsnt[0].dwOffset = 0;
	dsnt[0].hEventNotify = sound_nt_event;
	dsnt[1].dwOffset = DSOUND_BUFFER_HALF;
	dsnt[1].hEventNotify = sound_nt_event;

	lpdsnt->SetNotificationPositions(2, dsnt);
#endif

	// create critical section
	InitializeCriticalSection(&sound_nt_cs);

	// start new thread
	if ((sound_thread = CreateThread(NULL, 0, sound_proc, (LPVOID)this, 0, &sound_threadid)) == NULL) {
		out_logf(LOG_ERROR,_T("EMU::initialize_sound: CreateThread %ld"), GetLastError());
		return;
	}

	out_logf(LOG_INFO, _T("sound ok: rate:%d samples:%d latency:%d"), rate, samples, latency);
	sound_ok = true;
}
void EMU::initialize_sound(int rate, int samples)
{
	initialize_sound(rate, samples, 100);
}

void EMU::initialize_sound()
{
	// load sound config
#ifdef SUPPORT_SOUND_FREQ_55467HZ
	// PC-8801/9801 series
	static int freq_table[8] = {2000, 4000, 8000, 11025, 22050, 44100, 55467, 96000};
#else
	static int freq_table[8] = {2000, 4000, 8000, 11025, 22050, 44100, 48000, 96000};	// Hz
#endif
	static int late_table[5] = {50, 100, 200, 300, 400};	// ms

	if(!(0 <= config.sound_frequency && config.sound_frequency < 8)) {
		config.sound_frequency = 6;	// default: 48KHz
	}
	if(!(0 <= config.sound_latency && config.sound_latency < 5)) {
		config.sound_latency = 1;	// default: 100msec
	}
	int frequency = freq_table[config.sound_frequency];
	int samples = (int)((double)(frequency * late_table[config.sound_latency]) / 1000.0 + 0.5);
	int latency = late_table[config.sound_latency];

	initialize_sound(frequency, samples, latency);
}

void EMU::release_sound()
{
	// release direct sound
	if(lpdsnt) {
		lpdsnt->Release();
	}
	if(lpdsp) {
		lpdsp->Release();
	}
	if(lpdsb) {
		lpdsb->Release();
	}
	if(lpds) {
		lpds->Release();
	}
	lpdsnt = NULL;
	lpdsp = NULL;
	lpdsb = NULL;
	lpds = NULL;

	// stop recording
	stop_rec_sound();
}

void EMU::release_sound_on_emu_thread()
{
	// stop recording
	stop_rec_sound();
}

void EMU::start_sound()
{
	if(sound_ok) {
		// start play
		if(!sound_started) {
			lpdsb->Play(0, 0, DSBPLAY_LOOPING);
			sound_started = true;
		}
	}
}

void EMU::end_sound()
{
	if (sound_started) {
		sound_finished = true;
		// stop sound
		lpdsb->Stop();
	}
	if (sound_ok) {
		sound_ok = false;
#ifdef USE_SOUND_NOTIFY
		// close event
		if (sound_nt_event) {
			CloseHandle(sound_nt_event);
		}
#endif

		// wait thread
		EnterCriticalSection(&sound_nt_cs);
		LeaveCriticalSection(&sound_nt_cs);
		if (WaitForSingleObject(sound_thread, 10000) == WAIT_TIMEOUT) {
			// kill thread by force
			if (sound_thread) {
				TerminateThread(sound_thread, 0);
			}
		}
		// delete critical section
		DeleteCriticalSection(&sound_nt_cs);
	}
}

//void EMU::update_sound(int* extra_frames)
//{
//}

/// @attention called by another thread
DWORD EMU::update_sound_th()
{
//	out_debug(_T("EMU::update_sound_nt %ld"),timeGetTime());
	DWORD play_c, write_c, offset, size1, size2, next_wait;
	WORD *ptr1, *ptr2;

	// need finish?
	if (sound_finished) {
		return -1;
	}

	// check current playing position
	if(FAILED(lpdsb->GetCurrentPosition(&play_c, &write_c))) {
		out_log(LOG_ERROR,_T("EMU::update_sound_th GetCurrentPosition failed."));
		return -1;
	}
	if(play_c < DSOUND_BUFFER_HALF) {
		offset = DSOUND_BUFFER_HALF;
		next_wait = DSOUND_BUFFER_HALF + (DSOUND_BUFFER_HALF / 4) - play_c;
	} else {
		offset = 0;
		next_wait = DSOUND_BUFFER_SIZE + (DSOUND_BUFFER_HALF / 4) - play_c;
	}

	// sound buffer must be updated
	int extra_frames = 0;

	EnterCriticalSection(&sound_nt_cs);
	int16* sound_buffer = vm->create_sound(&extra_frames, DSOUND_SAMPLE);
	LeaveCriticalSection(&sound_nt_cs);

	if(lpdsb->Lock(offset, DSOUND_BUFFER_HALF, (void **)&ptr1, &size1, (void**)&ptr2, &size2, 0) == DSERR_BUFFERLOST) {
//		out_log(LOG_ERROR,_T("EMU::update_sound DSERR_BUFFERLOST"));
		lpdsb->Restore();
	}
	if(sound_buffer) {
		if(ptr1) {
			CopyMemory(ptr1, sound_buffer, size1);
		}
		if(ptr2) {
			CopyMemory(ptr2, sound_buffer + size1, size2);
		}
	}
	lpdsb->Unlock(ptr1, size1, ptr2, size2);
//	first_half = !first_half;

//	out_logf(LOG_DEBUG,_T("EMU::update_sound off:%ld playc:%ld writec:%ld size1:%ld size2:%ld next:%ld %dms"),offset,play_c,write_c,size1,size2,next_wait,(next_wait * 1000 / sound_rate / 4));

	return (next_wait * 1000 / sound_rate / 4);	// ms
}

DWORD WINAPI EMU::sound_proc(LPVOID lpParameter)
{
	EMU *emu = (EMU *)lpParameter;
	bool working = true;
	DWORD dwResult = 0;

#ifdef USE_SOUND_NOTIFY
	while(working){
		// wait event
		dwResult = MsgWaitForMultipleObjects(1, &emu->sound_nt_event, FALSE, INFINITE, QS_ALLEVENTS);

		switch(dwResult) {
		case WAIT_OBJECT_0 + 0:
			emu->update_sound_th();
			break;
		}
	}
#else
	while(working) {
		Sleep(dwResult);
		if ((dwResult = emu->update_sound_th()) == -1) {
			working = false;
			break;
		}
	}
#endif
	return 0;
}

void EMU::mute_sound(bool mute)
{
	if (mute) {
		now_mute = true;
	} else {
		if(!vm_pause) {
			now_mute = false;
#if 0
			if (sound_started) {
				lpdsb->Play(0, 0, DSBPLAY_LOOPING);
			}
#endif
		}
	}
}

void EMU::set_volume(int volume)
{
//	if (lpdsb) {
//		lpdsb->SetVolume((LONG)volume);
//	}
	// set volume for sound devices.
	if (vm) {
		vm->set_volume();
	}
}

uint32 EMU::adjust_sound_pos(uint32 msec)
{
	DWORD play_c, write_c;
	DWORD cur_sec = 0;

	while(cur_sec < (msec + 200)) {
		// check current playing position
		lpdsb->GetCurrentPosition(&play_c, &write_c);
		if ((DSOUND_BUFFER_HALF / 10) < play_c && play_c < (DSOUND_BUFFER_HALF * 4 / 10) && msec <= (cur_sec + 200)) {
			break;
		}
		Sleep(10);
		cur_sec += 10;
	}
	out_debugf(_T("EMU::adjust_sound: p:%ld %ldms"), play_c, cur_sec);
	return cur_sec;
}

void EMU::start_rec_sound()
{
	if(!now_rec_snd) {
		const _TCHAR *app_path;
		_TCHAR file_path[_MAX_PATH];
		app_path = _tcslen(config.snapshot_path) > 0 ? config.snapshot_path : application_path();
		_stprintf(file_path, _T("%s%04d-%02d-%02d_%02d-%02d-%02d.wav"), app_path, sTime.wYear, sTime.wMonth, sTime.wDay, sTime.wHour, sTime.wMinute, sTime.wSecond);

		rec = new FILEIO();
		if(rec->Fopen(file_path, FILEIO_WRITE_BINARY)) {
			// write dummy wave header
			wavheader_t header;

			memset(&header, 0, sizeof(wavheader_t));
			rec->Fwrite(&header, sizeof(wavheader_t), 1);
			rec_bytes = 0;
			now_rec_snd = true;
		}
		else {
			// failed to open the wave file
			out_logf(LOG_ERROR,_T("Failed to open %s."), file_path);
			delete rec;
		}
	}
}

void EMU::stop_rec_sound()
{
	if(now_rec_snd) {
		now_rec_snd = false;

		// update wave header
		wavheader_t header;

		header.dwRIFF = 0x46464952;
		header.dwFileSize = rec_bytes + sizeof(wavheader_t) - 8;
		header.dwWAVE = 0x45564157;
		header.dwfmt_ = 0x20746d66;
		header.dwFormatSize = 16;
		header.wFormatTag = 1;
		header.wChannels = 2;
		header.wBitsPerSample = 16;
		header.dwSamplesPerSec = sound_rate;
		header.wBlockAlign = header.wChannels * header.wBitsPerSample / 8;
		header.dwAvgBytesPerSec = header.dwSamplesPerSec * header.wBlockAlign;
		header.dwdata = 0x61746164;
		header.dwDataLength = rec_bytes;

		rec->Fseek(0, FILEIO_SEEK_SET);
		rec->Fwrite(&header, sizeof(wavheader_t), 1);
		rec->Fclose();

		delete rec;
	}
}

void EMU::restart_rec_sound()
{
	if(now_rec_snd) {
		rec->Fclose();
		delete rec;
		now_rec_snd = false;

		start_rec_sound();
	}
}

void EMU::record_rec_sound(uint8 *buffer, int samples)
{
	if(now_rec_snd) {
		// record sound
		int length = sizeof(uint8) * samples * 2;
		rec->Fwrite(buffer, length, 1);
		rec_bytes += length;
	}
}

void EMU::record_rec_sound(int16 *buffer, int samples)
{
	if(now_rec_snd) {
		// record sound
		int length = sizeof(int16) * samples * 2;
		rec->Fwrite(buffer, length, 1);
		rec_bytes += length;
	}
}

void EMU::record_rec_sound(int32 *buffer, int samples)
{
#ifdef SOUND_RECORD_IN_MIXER
	if(now_rec_snd) {
		// record sound
		int length = sizeof(uint16) * samples * 2;
		uint16 dat;
		for(int i=0; i<(samples << 1); i++) {
			dat = (uint16)buffer[i];
			rec->Fwrite(&dat, sizeof(uint16), 1);
		}
		rec_bytes += length;
	}
#endif
}

void EMU::lock_sound_buffer()
{
	EnterCriticalSection(&sound_nt_cs);
}
void EMU::unlock_sound_buffer()
{
	LeaveCriticalSection(&sound_nt_cs);
}
